################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
"""General-purpose utilities."""
import collections
import json
import os
import sys
import warnings
from contextlib import contextmanager
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import sympy

from .typing import AnyPath, DumpTarget, LoadSource

RNDSEED = 12345


def convert_dict_to_array(dictionary: dict) -> np.ndarray:
    """Convert a dictionary to a numpy array.

    Args:
        dictionary (dict): the dict containing the data

    Returns:
        array (numpy.array): a numpy array
    """

    array = np.array(dictionary["real"])

    if dictionary.get("imag"):
        array = array + 1j * np.array(dictionary["imag"])

    return array


def convert_array_to_dict(array: np.ndarray) -> dict:
    """Convert a numpy array to a dictionary.

    Args:
        array (numpy.array): a numpy array

    Returns:
        dictionary (dict): the dict containing the data
    """

    dictionary = {}
    if np.iscomplexobj(array):
        dictionary["real"] = array.real.tolist()
        dictionary["imag"] = array.imag.tolist()
    else:
        dictionary["real"] = array.tolist()

    return dictionary


def dec2bin(number: int, length: int) -> List[int]:
    """Converts a decimal number into a binary representation
    of fixed number of bits.

    Args:
        number: (int) the input decimal number
        length: (int) number of bits in the output string

    Returns:
        A list of binary numbers
    """

    if pow(2, length) < number:
        sys.exit(
            "Insufficient number of bits for representing the number {}".format(number)
        )

    bit_str = bin(number)
    bit_str = bit_str[2 : len(bit_str)]  # chop off the first two chars
    bit_string = [int(x) for x in list(bit_str)]
    if len(bit_string) < length:
        len_zeros = length - len(bit_string)
        bit_string = [int(x) for x in list(np.zeros(len_zeros))] + bit_string

    return bit_string


def bin2dec(x: List[int]) -> int:
    """Converts a binary vector to an integer, with the 0-th
    element being the most significant digit.

    Args:
        x: (list) a binary vector

    Returns:
        An integer
    """

    dec = 0
    coeff = 1
    for i in range(len(x)):
        dec = dec + coeff * x[len(x) - 1 - i]
        coeff = coeff * 2
    return dec


# The functions PAULI_X, PAULI_Y, PAULI_Z and IDENTITY below are used for
# generating the generators of the Pauli group, which include Pauli X, Y, Z
# operators as well as identity operator

pauli_x = np.array([[0.0, 1.0], [1.0, 0.0]])
pauli_y = np.array([[0.0, -1.0j], [1.0j, 0.0]])
pauli_z = np.array([[1.0, 0.0], [0.0, -1.0]])
identity = np.array([[1.0, 0.0], [0.0, 1.0]])


def is_identity(u: np.ndarray, tol=1e-15) -> bool:
    """Test if a matrix is identity.

    Args:
        u: np.ndarray
            Matrix to be checked.
        tol: float
            Threshold below which two matrix elements are considered equal.
    """

    dims = np.array(u).shape
    if dims[0] != dims[1]:
        raise Exception("Input matrix is not square.")

    return np.allclose(u, np.eye(u.shape[0]), atol=tol)


def is_unitary(u: np.ndarray, tol=1e-15) -> bool:
    """Test if a matrix is unitary.

    Args:
        u: array
            Matrix to be checked.
        tol: float
            Threshold below which two matrix elements are considered equal.
    """

    dims = np.array(u).shape
    if dims[0] != dims[1]:
        raise Exception("Input matrix is not square.")

    test_matrix = np.dot(np.array(u).T.conj(), u)
    return is_identity(test_matrix, tol)


def compare_unitary(u1: np.ndarray, u2: np.ndarray, tol: float = 1e-15) -> bool:
    """Compares two unitary operators to see if they are equal to within a phase.

    Args:
        u1 (numpy.ndarray): First unitary operator.
        u2 (numpy.ndarray): Second unitary operator.
        tol (float): Threshold below which two matrix elements are considered equal.

    Returns:
        bool: True if the unitaries are equal to within the tolerance, ignoring
            differences in global phase.
    """

    if not is_unitary(u1, tol):
        raise Exception("The first input matrix is not unitary.")
    if not is_unitary(u2, tol):
        raise Exception("The second input matrix is not unitary.")

    test_matrix = np.dot(u1.conj().T, u2)
    phase = test_matrix.item((0, 0)) ** -1
    return is_identity(phase * test_matrix, tol)


def sample_from_probability_distribution(
    probability_distribution: dict, n_samples: int
) -> collections.Counter:
    """
    Samples events from a discrete probability distribution

    Args:
        probabilty_distribution: The discrete probability distribution to be used
        for sampling. This should be a dictionary

        n_samples (int): The number of samples desired

    Returns:
        A dictionary of the outcomes sampled. The key values are the things be sampled
        and values are how many times those things appeared in the sampling
    """
    if isinstance(probability_distribution, dict):
        # Need to do this preprocessing to handle different types of dict keys
        keys_as_array = np.empty(len(probability_distribution), dtype=object)
        keys_as_array[:] = list(probability_distribution.keys())

        result = np.random.choice(
            keys_as_array,
            n_samples,
            replace=True,
            p=list(probability_distribution.values()),
        )
        sampled_dict: collections.Counter = collections.Counter(result)

        return sampled_dict
    else:
        raise RuntimeError(
            "Probability distribution should be a dictionary with key value \
        being the thing being sampled and the value being probability of getting \
        sampled "
        )


def convert_bitstrings_to_tuples(bitstrings: Iterable[str]) -> List[Tuple[int, ...]]:
    """Given the measured bitstrings, convert each bitstring to tuple format

    Args:
        bitstrings (list of strings): the measured bitstrings
    Returns:
        A list of tuples
    """
    # Convert from bitstrings to tuple format
    measurements: List[Tuple[int, ...]] = []
    for bitstring in bitstrings:

        measurement: Tuple[int, ...] = ()
        for char in bitstring:
            measurement = measurement + (int(char),)

        measurements.append(measurement)
    return measurements


def convert_tuples_to_bitstrings(tuples: List[Tuple[int]]) -> List[str]:
    """Given a set of measurement tuples, convert each to a little endian
    string.

    Args:
        tuples (list of tuples): the measurement tuples
    Returns:
        A list of bitstrings
    """
    # Convert from tuples to bitstrings
    bitstrings = ["".join(map(str, tup)) for tup in tuples]
    return bitstrings


class ValueEstimate(float):
    """A class representing a numerical value and its precision corresponding
        to an observable or an objective function

    Args:
        value (np.float): the numerical value or a value that can be converted to float
        precision (np.float): its precision

    Attributes:
        value (np.float): the numerical value
        precision (np.float): its precision
    """

    def __init__(self, value, precision: Optional[float] = None):
        super().__init__()
        self.precision = precision

    def __new__(cls, value, precision=None):
        return super().__new__(cls, value)

    def __eq__(self, other):
        super_eq = super().__eq__(other)
        if super_eq is NotImplemented:
            return super_eq
        return super_eq and self.precision == getattr(other, "precision", None)

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        value_str = super().__str__()
        if self.precision is not None:
            return f"{value_str} ± {self.precision}"
        else:
            return f"{value_str}"

    def to_dict(self):
        """Convert to a dictionary"""

        data = {}
        if type(self).__module__ == np.__name__:
            data["value"] = self.value.item()
        else:
            data["value"] = self

        if type(self.precision).__module__ == np.__name__:
            data["precision"] = self.precision.item()
        else:
            data["precision"] = self.precision

        return data

    @classmethod
    def from_dict(cls, dictionary):
        """Create an ExpectationValues object from a dictionary."""

        value = dictionary["value"]
        if "precision" in dictionary:
            precision = dictionary["precision"]
            return cls(value, precision)
        else:
            return cls(value)


def load_value_estimate(file: LoadSource) -> ValueEstimate:
    """Loads value estimate from a failed.

    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        array (numpy.array): the array
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)  # type: ignore

    return ValueEstimate.from_dict(data)


def save_value_estimate(value_estimate: ValueEstimate, filename: AnyPath):
    """Saves value estimate to a file.

    Args:
        value_estimate (orquestra.quantum.utils.ValueEstimate): the value estimate
        file (str or file-like object): the name of the file, or a file-like object
    """
    dictionary = value_estimate.to_dict()

    with open(filename, "w") as f:
        f.write(json.dumps(dictionary, indent=2))


def load_list(file: LoadSource) -> List:
    """Load an array from a file.

    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        array (list): the list
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)  # type: ignore

    return data["list"]


def save_list(array: List, filename: AnyPath):
    """Save expectation values to a file.

    Args:
        array (list): the list to be saved
        file (str or file-like object): the name of the file, or a file-like object
    """
    dictionary: Dict[str, Any] = {}
    dictionary["list"] = array

    with open(filename, "w") as f:
        f.write(json.dumps(dictionary, indent=2))


def save_generic_dict(dictionary: Dict, filename: AnyPath):
    """Save dictionary as json

    Args:
        dictionary (dict): the dict containing the data
    """
    dictionary_stored = {}
    dictionary_stored.update(dictionary)

    with open(filename, "w") as f:
        f.write(json.dumps(dictionary_stored, indent=2))


def create_symbols_map(
    symbols: List[sympy.Symbol], params: np.ndarray
) -> Dict[sympy.Symbol, float]:
    """
    Creates a map to be used for evaluating sympy expressions.

    Args:
        symbols: list of sympy Symbols to be evaluated
        params: numpy array containing numerical value for the symbols
    """
    if len(symbols) != len(params):
        raise (
            ValueError(
                "Length of symbols: {0} doesn't match length of params: {1}".format(
                    len(symbols), len(params)
                )
            )
        )
    return {symbol: param for symbol, param in zip(symbols, params.tolist())}


def save_timing(walltime: float, filename: AnyPath) -> None:
    """
    Saves timing information.

    Args:
        walltime: The execution time.
    """

    with open(filename, "w") as f:
        f.write(json.dumps({"walltime": walltime}))


def save_nmeas_estimate(
    nmeas: float, nterms: int, filename: AnyPath, frame_meas: np.ndarray = None
) -> None:
    """Save an estimate of the number of measurements to a file

    Args:
        nmeas: total number of measurements for epsilon = 1.0
        nterms: number of terms (groups) in the objective function
        frame_meas: A list of the number of measurements per frame for epsilon = 1.0
    """

    data: Dict[str, Any] = {}
    data["K"] = nmeas
    data["nterms"] = nterms
    if frame_meas is not None:
        data["frame_meas"] = convert_array_to_dict(frame_meas)

    with open(filename, "w") as f:
        f.write(json.dumps(data, indent=2))


def load_nmeas_estimate(filename: AnyPath) -> Tuple[float, int, np.ndarray]:
    """Load an estimate of the number of measurements from a file.

    Args:
        filename: the name of the file

    Returns:
        nmeas: number of measurements for epsilon = 1.0
        nterms: number of terms in the hamiltonian
        frame_meas: frame measurements (number of measurements per group)
    """

    with open(filename, "r") as f:
        data = json.load(f)

    frame_meas = convert_dict_to_array(data["frame_meas"])
    K_coeff = data["K"]
    nterms = data["nterms"]

    return K_coeff, nterms, frame_meas


def scale_and_discretize(values: Iterable[float], total: int) -> List[int]:
    """Convert a list of floats to a list of integers such that the total equals
    a given value and the ratios of elements are approximately preserved.

    Args:
        values: The list of floats to be scaled and discretized.
        total: The desired total which the resulting values should sum to.

    Returns:
        A list of integers whose sum is equal to the given total, where the
            ratios of the list elements are approximately equal to the ratios
            of the input list elements.
    """

    # Note: combining the two lines below breaks type checking; see mypy #6040
    value_sum = sum(values)
    scale_factor = total / value_sum

    result = [np.floor(value * scale_factor) for value in values]
    remainders = [
        value * scale_factor - np.floor(value * scale_factor) for value in values
    ]
    indexes_sorted_by_remainder = np.argsort(remainders)[::-1]
    for index in range(int(round(total - sum(result)))):
        result[indexes_sorted_by_remainder[index]] += 1

    result = [int(value) for value in result]

    assert sum(result) == total, "The scaled list does not sum to the desired total."

    return result


def get_ordered_list_of_bitstrings(num_qubits: int) -> List[str]:
    """Create list of binary strings corresponding to 2^num_qubits integers
    and save them in ascending order.

    Args:
        num_qubits: number of binary digits in each bitstring

    Returns:
        The ordered bitstring representations of the integers
    """
    bitstrings = []
    for i in range(2**num_qubits):
        bitstring = "{0:b}".format(i)
        while len(bitstring) < num_qubits:
            bitstring = "0" + bitstring
        bitstrings.append(bitstring)
    return bitstrings


@contextmanager
def ensure_open(path_like: Union[LoadSource, DumpTarget], mode="r", encoding="utf-8"):
    # str | bytes | PathLike | Readable
    if isinstance(path_like, (str, bytes, os.PathLike)):
        with open(path_like, mode, encoding=encoding if "b" not in mode else None) as f:
            yield f
    else:
        # Readable | Writable
        if set(mode).intersection(set("wxa+")) and not path_like.writable():
            raise ValueError(f"File isn't writable, can't ensure mode {mode}")
        yield path_like



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""testing angular momentum generators. _fermion_spin_operators.py"""
import unittest

from orquestra.quantum.openfermion.hamiltonians.special_operators import number_operator
from orquestra.quantum.openfermion.ops.operators import FermionOperator


class NumberOperatorTest(unittest.TestCase):
    def test_fermion_number_operator_site(self):
        op = number_operator(3, 2, 1j, -1)
        self.assertEqual(op, FermionOperator(((2, 1), (2, 0))) * 1j)

    def test_number_operator_nosite(self):
        op = number_operator(4, parity=-1)
        expected = (
            FermionOperator(((0, 1), (0, 0)))
            + FermionOperator(((1, 1), (1, 0)))
            + FermionOperator(((2, 1), (2, 0)))
            + FermionOperator(((3, 1), (3, 0)))
        )
        self.assertEqual(op, expected)

    def test_bad_parity(self):
        with self.assertRaises(ValueError):
            number_operator(4, parity=2)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest
import sympy

from orquestra.quantum.api.backend import QuantumSimulator
from orquestra.quantum.api.backend_test import (
    QuantumSimulatorGatesTest,
    QuantumSimulatorTests,
)
from orquestra.quantum.circuits import (
    CNOT,
    RX,
    RY,
    U3,
    XX,
    Circuit,
    GateOperation,
    MultiPhaseOperation,
    Operation,
)
from orquestra.quantum.symbolic_simulator import SymbolicSimulator


@pytest.fixture
def backend():
    return SymbolicSimulator()


@pytest.fixture
def wf_simulator() -> SymbolicSimulator:
    return SymbolicSimulator()


class SymbolicSimulatorWithNonSupportedOperations(SymbolicSimulator):
    def is_natively_supported(self, operation: Operation) -> bool:
        if isinstance(operation, GateOperation) and operation.gate.name == "RX":
            return False
        return super().is_natively_supported(operation)


class SymbolicSimulatorWithDefaultSetOfSupportedOperations(SymbolicSimulator):
    def is_natively_supported(self, operation: Operation) -> bool:
        return QuantumSimulator.is_natively_supported(self, operation)


class TestSymbolicSimulator(QuantumSimulatorTests):
    gates_list = [
        XX(sympy.Symbol("theta"))(2, 1),
        U3(
            sympy.Symbol("alpha"),
            sympy.Symbol("beta"),
            sympy.Symbol("gamma"),
        )(1),
    ]
    incorrect_bindings = [
        {},
        {
            "alpha": sympy.pi,
            "beta": sympy.pi,
        },
    ]
    correct_bindings = [
        dict(zip(gate.free_symbols, [sympy.pi] * len(tuple(gate.free_symbols))))
        for gate in gates_list
    ]

    @pytest.mark.parametrize(
        "gate, binding",
        list(zip(gates_list, incorrect_bindings)),
    )
    def test_cannot_sample_from_circuit_containing_free_symbols(
        self, wf_simulator, gate, binding
    ):

        circuit = Circuit([gate])
        with pytest.raises(ValueError):
            wf_simulator.run_circuit_and_measure(
                circuit, n_samples=1000, symbol_map=binding
            )

    @pytest.mark.parametrize(
        "gate, binding",
        list(zip(gates_list, correct_bindings)),
    )
    def test_passes_for_complete_bindings(self, wf_simulator, gate, binding):
        circuit = Circuit([gate])
        wf_simulator.run_circuit_and_measure(
            circuit, n_samples=1000, symbol_map=binding
        )

    @pytest.mark.parametrize(
        "circuit",
        [
            Circuit([RY(0.5)(0), RX(1)(1), CNOT(0, 2), RX(np.pi)(2)]),
            Circuit([RX(1)(1), CNOT(0, 2), RX(np.pi)(2), RY(0.5)(0)]),
            Circuit([RX(1)(1), CNOT(0, 2), RX(np.pi)(2), RY(0.5)(0), RX(0.5)(0)]),
        ],
    )
    def test_quantum_simulator_switches_between_native_and_nonnative_modes_of_execution(
        self, circuit
    ):
        simulator = SymbolicSimulatorWithNonSupportedOperations()
        reference_simulator = SymbolicSimulator()

        np.testing.assert_array_equal(
            simulator.get_wavefunction(circuit).amplitudes,
            reference_simulator.get_wavefunction(circuit).amplitudes,
        )

    def test_by_default_only_gate_operations_are_supported(self):
        simulator = SymbolicSimulatorWithDefaultSetOfSupportedOperations()
        assert simulator.is_natively_supported(RX(np.pi / 2)(1))
        assert not simulator.is_natively_supported(
            MultiPhaseOperation((0.5, 0.2, 0.3, 0.1))
        )


class TestSymbolicSimulatorGates(QuantumSimulatorGatesTest):
    pass



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for operator_utils."""

import itertools
import os
import unittest

import numpy
import sympy
from scipy.sparse import csc_matrix

from orquestra.quantum.openfermion.config import DATA_DIRECTORY
from orquestra.quantum.openfermion.hamiltonians import fermi_hubbard
from orquestra.quantum.openfermion.ops.operators import (
    FermionOperator,
    IsingOperator,
    QubitOperator,
)
from orquestra.quantum.openfermion.ops.representations import InteractionOperator
from orquestra.quantum.openfermion.testing.testing_utils import (
    random_interaction_operator,
)
from orquestra.quantum.openfermion.transforms.opconversions import (
    bravyi_kitaev,
    jordan_wigner,
)
from orquestra.quantum.openfermion.transforms.repconversions import (
    get_interaction_operator,
)
from orquestra.quantum.openfermion.utils.operator_utils import (
    OperatorUtilsError,
    count_qubits,
    get_file_path,
    hermitian_conjugated,
    is_hermitian,
    is_identity,
    load_operator,
    save_operator,
)


class OperatorUtilsTest(unittest.TestCase):
    def setUp(self):
        self.n_qubits = 5
        self.fermion_term = FermionOperator("1^ 2^ 3 4", -3.17)
        self.fermion_operator = self.fermion_term + hermitian_conjugated(
            self.fermion_term
        )
        self.qubit_operator = jordan_wigner(self.fermion_operator)
        self.interaction_operator = get_interaction_operator(self.fermion_operator)
        self.ising_operator = IsingOperator("[Z0] + [Z1] + [Z2] + [Z3] + [Z4]")

    def test_n_qubits_single_fermion_term(self):
        self.assertEqual(self.n_qubits, count_qubits(self.fermion_term))

    def test_n_qubits_fermion_operator(self):
        self.assertEqual(self.n_qubits, count_qubits(self.fermion_operator))

    def test_n_qubits_qubit_operator(self):
        self.assertEqual(self.n_qubits, count_qubits(self.qubit_operator))

    def test_n_qubits_interaction_operator(self):
        self.assertEqual(self.n_qubits, count_qubits(self.interaction_operator))

    def test_n_qubits_ising_operator(self):
        self.assertEqual(self.n_qubits, count_qubits(self.ising_operator))

    def test_n_qubits_bad_type(self):
        with self.assertRaises(TypeError):
            count_qubits("twelve")

    def test_is_identity_unit_fermionoperator(self):
        self.assertTrue(is_identity(FermionOperator(())))

    def test_is_identity_double_of_unit_fermionoperator(self):
        self.assertTrue(is_identity(2.0 * FermionOperator(())))

    def test_is_identity_unit_qubitoperator(self):
        self.assertTrue(is_identity(QubitOperator(())))

    def test_is_identity_double_of_unit_qubitoperator(self):
        self.assertTrue(is_identity(QubitOperator((), 2.0)))

    def test_not_is_identity_single_term_fermionoperator(self):
        self.assertFalse(is_identity(FermionOperator("1^")))

    def test_not_is_identity_single_term_qubitoperator(self):
        self.assertFalse(is_identity(QubitOperator("X1")))

    def test_not_is_identity_zero_qubitoperator(self):
        self.assertFalse(is_identity(QubitOperator()))

    def test_is_identity_bad_type(self):
        with self.assertRaises(TypeError):
            _ = is_identity("eleven")


class HermitianConjugatedTest(unittest.TestCase):
    def test_hermitian_conjugated_qubit_op(self):
        """Test conjugating QubitOperators."""
        op = QubitOperator()
        op_hc = hermitian_conjugated(op)
        correct_op = op
        self.assertEqual(op_hc, correct_op)

        op = QubitOperator("X0 Y1", 2.0)
        op_hc = hermitian_conjugated(op)
        correct_op = op
        self.assertEqual(op_hc, correct_op)

        op = QubitOperator("X0 Y1", 2.0j)
        op_hc = hermitian_conjugated(op)
        correct_op = QubitOperator("X0 Y1", -2.0j)
        self.assertEqual(op_hc, correct_op)

        op = QubitOperator("X0 Y1", 2.0) + QubitOperator("Z4 X5 Y7", 3.0j)
        op_hc = hermitian_conjugated(op)
        correct_op = QubitOperator("X0 Y1", 2.0) + QubitOperator("Z4 X5 Y7", -3.0j)
        self.assertEqual(op_hc, correct_op)

    def test_hermitian_conjugated_qubit_op_consistency(self):
        """Some consistency checks for conjugating QubitOperators."""
        ferm_op = (
            FermionOperator("1^ 2")
            + FermionOperator("2 3 4")
            + FermionOperator("2^ 7 9 11^")
        )

        # Check that hermitian conjugation commutes with transforms
        self.assertEqual(
            jordan_wigner(hermitian_conjugated(ferm_op)),
            hermitian_conjugated(jordan_wigner(ferm_op)),
        )
        self.assertEqual(
            bravyi_kitaev(hermitian_conjugated(ferm_op)),
            hermitian_conjugated(bravyi_kitaev(ferm_op)),
        )

    def test_hermitian_conjugate_empty(self):
        op = FermionOperator()
        op = hermitian_conjugated(op)
        self.assertEqual(op, FermionOperator())

    def test_hermitian_conjugate_simple(self):
        op = FermionOperator("1^")
        op_hc = FermionOperator("1")
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

    def test_hermitian_conjugate_complex_const(self):
        op = FermionOperator("1^ 3", 3j)
        op_hc = -3j * FermionOperator("3^ 1")
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

    def test_hermitian_conjugate_notordered(self):
        op = FermionOperator("1 3^ 3 3^", 3j)
        op_hc = -3j * FermionOperator("3 3^ 3 1^")
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

    def test_hermitian_conjugate_semihermitian(self):
        op = (
            FermionOperator()
            + 2j * FermionOperator("1^ 3")
            + FermionOperator("3^ 1") * -2j
            + FermionOperator("2^ 2", 0.1j)
        )
        op_hc = (
            FermionOperator()
            + FermionOperator("1^ 3", 2j)
            + FermionOperator("3^ 1", -2j)
            + FermionOperator("2^ 2", -0.1j)
        )
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

    def test_hermitian_conjugated_empty(self):
        op = FermionOperator()
        self.assertEqual(op, hermitian_conjugated(op))

    def test_hermitian_conjugated_simple(self):
        op = FermionOperator("0")
        op_hc = FermionOperator("0^")
        self.assertEqual(op_hc, hermitian_conjugated(op))

    def test_hermitian_conjugated_complex_const(self):
        op = FermionOperator("2^ 2", 3j)
        op_hc = FermionOperator("2^ 2", -3j)
        self.assertEqual(op_hc, hermitian_conjugated(op))

    def test_hermitian_conjugated_multiterm(self):
        op = FermionOperator("1^ 2") + FermionOperator("2 3 4")
        op_hc = FermionOperator("2^ 1") + FermionOperator("4^ 3^ 2^")
        self.assertEqual(op_hc, hermitian_conjugated(op))

    def test_hermitian_conjugated_semihermitian(self):
        op = (
            FermionOperator()
            + 2j * FermionOperator("1^ 3")
            + FermionOperator("3^ 1") * -2j
            + FermionOperator("2^ 2", 0.1j)
        )
        op_hc = (
            FermionOperator()
            + FermionOperator("1^ 3", 2j)
            + FermionOperator("3^ 1", -2j)
            + FermionOperator("2^ 2", -0.1j)
        )
        self.assertEqual(op_hc, hermitian_conjugated(op))

    def test_hermitian_conjugated_interaction_operator(self):
        for n_orbitals, _ in itertools.product((1, 2, 5), range(5)):
            operator = random_interaction_operator(n_orbitals)
            qubit_operator = jordan_wigner(operator)
            conjugate_operator = hermitian_conjugated(operator)
            conjugate_qubit_operator = jordan_wigner(conjugate_operator)
            assert hermitian_conjugated(qubit_operator) == conjugate_qubit_operator

    def test_exceptions(self):
        with self.assertRaises(TypeError):
            _ = is_hermitian("a")

        with self.assertRaises(TypeError):
            _ = hermitian_conjugated(1)


class IsHermitianTest(unittest.TestCase):
    def test_fermion_operator_zero(self):
        op = FermionOperator()
        self.assertTrue(is_hermitian(op))

    def test_fermion_operator_identity(self):
        op = FermionOperator(())
        self.assertTrue(is_hermitian(op))

    def test_fermion_operator_nonhermitian(self):
        op = FermionOperator("0^ 1 2^ 3")
        self.assertFalse(is_hermitian(op))

    def test_fermion_operator_hermitian(self):
        op = FermionOperator("0^ 1 2^ 3")
        op += FermionOperator("3^ 2 1^ 0")
        self.assertTrue(is_hermitian(op))

        op = fermi_hubbard(2, 2, 1.0, 1.0)
        self.assertTrue(is_hermitian(op))

        # TODO: insert bose_hubbard here
        # op = fermi_hubbard(2, 2, 1., 1.)
        # self.assertTrue(is_hermitian(op))

    def test_qubit_operator_zero(self):
        op = QubitOperator()
        self.assertTrue(is_hermitian(op))

    def test_qubit_operator_identity(self):
        op = QubitOperator(())
        self.assertTrue(is_hermitian(op))

    def test_qubit_operator_nonhermitian(self):
        op = QubitOperator("X0 Y2 Z5", 1.0 + 2.0j)
        self.assertFalse(is_hermitian(op))

    def test_qubit_operator_hermitian(self):
        op = QubitOperator("X0 Y2 Z5", 1.0 + 2.0j)
        op += QubitOperator("X0 Y2 Z5", 1.0 - 2.0j)
        self.assertTrue(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_zero(self):
        op = numpy.zeros((4, 4))
        self.assertTrue(is_hermitian(op))
        op = csc_matrix(op)
        self.assertTrue(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_identity(self):
        op = numpy.eye(4)
        self.assertTrue(is_hermitian(op))
        op = csc_matrix(op)
        self.assertTrue(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_nonhermitian(self):
        op = numpy.arange(16).reshape((4, 4))
        self.assertFalse(is_hermitian(op))
        op = csc_matrix(op)
        self.assertFalse(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_hermitian(self):
        op = numpy.arange(16, dtype=complex).reshape((4, 4))
        op += 1.0j * op
        op += op.T.conj()
        self.assertTrue(is_hermitian(op))
        op = csc_matrix(op)
        self.assertTrue(is_hermitian(op))

    def test_exceptions(self):
        with self.assertRaises(TypeError):
            _ = is_hermitian("a")


class SaveLoadOperatorTest(unittest.TestCase):
    def setUp(self):
        self.n_qubits = 5
        self.fermion_term = FermionOperator("1^ 2^ 3 4", -3.17)
        self.fermion_operator = self.fermion_term + hermitian_conjugated(
            self.fermion_term
        )
        self.qubit_operator = jordan_wigner(self.fermion_operator)
        self.file_name = "test_file"

        self.bad_operator_filename = "bad_file.data"
        bad_op = "A:\nB"
        with open(os.path.join(DATA_DIRECTORY, self.bad_operator_filename), "w") as fid:
            fid.write(bad_op)

    def tearDown(self):
        file_path = os.path.join(DATA_DIRECTORY, self.file_name + ".data")
        if os.path.isfile(file_path):
            os.remove(file_path)
        file_path = os.path.join(DATA_DIRECTORY, self.bad_operator_filename)
        if os.path.isfile(file_path):
            os.remove(file_path)

    def test_save_sympy_plaintext(self):
        operator = FermionOperator("1^", sympy.Symbol("x"))
        with self.assertRaises(TypeError):
            save_operator(operator, self.file_name, plain_text=True)

    def test_raises_error_sympy(self):
        operator = FermionOperator("1^", sympy.Symbol("x"))
        with self.assertRaises(TypeError):
            save_operator(operator, self.file_name, plain_text=False)

    def test_save_and_load_fermion_operators(self):
        save_operator(self.fermion_operator, self.file_name)
        loaded_fermion_operator = load_operator(self.file_name)
        self.assertEqual(
            self.fermion_operator,
            loaded_fermion_operator,
            msg=str(self.fermion_operator - loaded_fermion_operator),
        )

    def test_save_and_load_fermion_operators_readably(self):
        save_operator(self.fermion_operator, self.file_name, plain_text=True)
        loaded_fermion_operator = load_operator(self.file_name, plain_text=True)
        self.assertTrue(self.fermion_operator == loaded_fermion_operator)

    def test_save_and_load_qubit_operators(self):
        save_operator(self.qubit_operator, self.file_name)
        loaded_qubit_operator = load_operator(self.file_name)
        self.assertTrue(self.qubit_operator == loaded_qubit_operator)

    def test_save_and_load_qubit_operators_readably(self):
        save_operator(self.qubit_operator, self.file_name, plain_text=True)
        loaded_qubit_operator = load_operator(self.file_name, plain_text=True)
        self.assertEqual(self.qubit_operator, loaded_qubit_operator)

    def test_load_bad_operator(self):
        with self.assertRaises(TypeError):
            load_operator(self.bad_operator_filename, plain_text=True)

    def test_save_readably(self):
        save_operator(self.fermion_operator, self.file_name, plain_text=True)
        file_path = os.path.join(DATA_DIRECTORY, self.file_name + ".data")
        with open(file_path, "r") as f:
            self.assertEqual(
                f.read(),
                "\n".join(
                    ["FermionOperator:", "-3.17 [1^ 2^ 3 4] +", "-3.17 [4^ 3^ 2 1]"]
                ),
            )

    def test_save_no_filename_operator_utils_error(self):
        with self.assertRaises(OperatorUtilsError):
            save_operator(self.fermion_operator)

    def test_basic_save(self):
        save_operator(self.fermion_operator, self.file_name)

    def test_save_interaction_operator_not_implemented(self):
        constant = 100.0
        one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float
        )
        one_body[1, 1] = 10.0
        two_body[1, 2, 3, 4] = 12.0
        interaction_operator = InteractionOperator(constant, one_body, two_body)
        with self.assertRaises(NotImplementedError):
            save_operator(interaction_operator, self.file_name)

    def test_save_on_top_of_existing_operator_utils_error(self):
        save_operator(self.fermion_operator, self.file_name)
        with self.assertRaises(OperatorUtilsError):
            save_operator(self.fermion_operator, self.file_name)

    def test_save_on_top_of_existing_operator_error_with_explicit_flag(self):
        save_operator(self.fermion_operator, self.file_name)
        with self.assertRaises(OperatorUtilsError):
            save_operator(self.fermion_operator, self.file_name, allow_overwrite=False)

    def test_overwrite_flag_save_on_top_of_existing_operator(self):
        save_operator(self.fermion_operator, self.file_name)
        save_operator(self.fermion_operator, self.file_name, allow_overwrite=True)
        fermion_operator = load_operator(self.file_name)

        self.assertEqual(fermion_operator, self.fermion_operator)

    def test_load_bad_type(self):
        with self.assertRaises(TypeError):
            _ = load_operator("bad_type_operator")

    def test_save_bad_type(self):
        with self.assertRaises(TypeError):
            save_operator("ping", "somewhere")


class GetFileDirTest(unittest.TestCase):
    def setUp(self):
        self.filename = "foo"
        self.datadirname = "data"

    def test_file_load(self):
        """Test if file name is acquired correctly"""
        filepath = get_file_path(self.filename, None)
        self.assertEqual(filepath, DATA_DIRECTORY + "/" + self.filename + ".data")

        filepath = get_file_path(self.filename, self.datadirname)
        self.assertEqual(filepath, self.datadirname + "/" + self.filename + ".data")



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Bravyi-Kitaev transform on fermionic operators."""

from orquestra.quantum.openfermion.ops.operators import FermionOperator, QubitOperator
from orquestra.quantum.openfermion.ops.representations import InteractionOperator
from orquestra.quantum.openfermion.utils.operator_utils import count_qubits


def bravyi_kitaev(operator, n_qubits=None):
    """Apply the Bravyi-Kitaev transform.

    Implementation from arXiv:quant-ph/0003137 and
    "A New Data Structure for Cumulative Frequency Tables" by Peter M. Fenwick.

    Note that this implementation is equivalent to the one described in
    arXiv:1208.5986, and is different from the one described in
    arXiv:1701.07072. The one described in arXiv:1701.07072 is implemented
    in OpenFermion as `bravyi_kitaev_tree`.

    Args:
        operator (openfermion.ops.FermionOperator):
            A FermionOperator to transform.
        n_qubits (int|None):
            Can force the number of qubits in the resulting operator above the
            number that appear in the input operator.

    Returns:
        transformed_operator: An instance of the QubitOperator class.

    Raises:
        ValueError: Invalid number of qubits specified.
    """
    if isinstance(operator, FermionOperator):
        return _bravyi_kitaev_fermion_operator(operator, n_qubits)
    if isinstance(operator, InteractionOperator):
        return _bravyi_kitaev_interaction_operator(operator, n_qubits)
    raise TypeError(
        "Couldn't apply the Bravyi-Kitaev Transform to object "
        "of type {}.".format(type(operator))
    )


def _update_set(index, n_qubits):
    """The bits that need to be updated upon flipping the occupancy
    of a mode."""
    indices = set()

    # For bit manipulation we need to count from 1 rather than 0
    index += 1
    # Ensure index is not a member of the set
    index += index & -index

    while index <= n_qubits:
        indices.add(index - 1)
        # Add least significant one to index
        # E.g. 00010100 -> 00011000
        index += index & -index
    return indices


def _occupation_set(index):
    """The bits whose parity stores the occupation of mode `index`."""
    indices = set()

    # For bit manipulation we need to count from 1 rather than 0
    index += 1

    indices.add(index - 1)
    parent = index & (index - 1)
    index -= 1
    while index != parent:
        indices.add(index - 1)
        # Remove least significant one from index
        # E.g. 00010100 -> 00010000
        index &= index - 1
    return indices


def _parity_set(index):
    """The bits whose parity stores the parity of the bits 0 .. `index`."""
    indices = set()

    while index > 0:
        indices.add(index - 1)
        # Remove least significant one from index
        # E.g. 00010100 -> 00010000
        index &= index - 1
    return indices


def _bravyi_kitaev_majorana_operator(operator, n_qubits):
    # Compute the number of qubits.
    N = count_qubits(operator)
    if n_qubits is None:
        n_qubits = N
    if n_qubits < N:
        raise ValueError("Invalid number of qubits specified.")

    # Compute transformed operator.
    transformed_terms = (
        _transform_majorana_term(term=term, coefficient=coeff, n_qubits=n_qubits)
        for term, coeff in operator.terms.items()
    )
    return inline_sum(summands=transformed_terms, seed=QubitOperator())


def _transform_majorana_term(term, coefficient, n_qubits):
    # Build the Bravyi-Kitaev transformed operators.
    transformed_ops = (
        _transform_majorana_operator(majorana_index, n_qubits)
        for majorana_index in term
    )
    return inline_product(factors=transformed_ops, seed=QubitOperator((), coefficient))


def _transform_majorana_operator(majorana_index, n_qubits):
    q, b = divmod(majorana_index, 2)

    update_set = _update_set(q, n_qubits)
    update_set.add(q)
    occupation_set = _occupation_set(q)
    parity_set = _parity_set(q)

    if b:
        return QubitOperator(
            [(q, "Y")]
            + [(i, "X") for i in update_set - {q}]
            + [(i, "Z") for i in (parity_set ^ occupation_set) - {q}]
        )
    else:
        return QubitOperator(
            [(i, "X") for i in update_set] + [(i, "Z") for i in parity_set]
        )


def _transform_operator_term(term, coefficient, n_qubits):
    """
    Args:
        term (list[tuple[int, int]]):
            A list of (mode, raising-vs-lowering) ladder operator terms.
        coefficient (float):
        n_qubits (int):
    Returns:
        QubitOperator:
    """

    # Build the Bravyi-Kitaev transformed operators.
    transformed_ladder_ops = (
        _transform_ladder_operator(ladder_operator, n_qubits)
        for ladder_operator in term
    )
    return inline_product(
        factors=transformed_ladder_ops, seed=QubitOperator((), coefficient)
    )


def _bravyi_kitaev_fermion_operator(operator, n_qubits):
    # Compute the number of qubits.
    N = count_qubits(operator)
    if n_qubits is None:
        n_qubits = N
    if n_qubits < N:
        raise ValueError("Invalid number of qubits specified.")

    # Compute transformed operator.
    transformed_terms = (
        _transform_operator_term(
            term=term, coefficient=operator.terms[term], n_qubits=n_qubits
        )
        for term in operator.terms
    )
    return inline_sum(summands=transformed_terms, seed=QubitOperator())


def _transform_ladder_operator(ladder_operator, n_qubits):
    """
    Args:
        ladder_operator (tuple[int, int]): the ladder operator
        n_qubits (int): the number of qubits
    Returns:
        QubitOperator
    """
    index, action = ladder_operator

    update_set = _update_set(index, n_qubits)
    update_set.add(index)
    occupation_set = _occupation_set(index)
    parity_set = _parity_set(index)

    # Initialize the transformed majorana operator (a_p^\dagger + a_p) / 2
    transformed_operator = QubitOperator(
        [(i, "X") for i in update_set] + [(i, "Z") for i in parity_set], 0.5
    )
    # Get the transformed (a_p^\dagger - a_p) / 2
    # Below is equivalent to X(update_set) * Z(parity_set ^ occupation_set)
    transformed_majorana_difference = QubitOperator(
        [(index, "Y")]
        + [(i, "X") for i in update_set - {index}]
        + [(i, "Z") for i in (parity_set ^ occupation_set) - {index}],
        -0.5j,
    )

    # Raising
    if action == 1:
        transformed_operator += transformed_majorana_difference
    # Lowering
    else:
        transformed_operator -= transformed_majorana_difference

    return transformed_operator


def inline_sum(summands, seed):
    """Computes a sum, using the __iadd__ operator.
    Args:
        seed (T): The starting total. The zero value.
        summands (iterable[T]): Values to add (with +=) into the total.
    Returns:
        T: The result of adding all the factors into the zero value.
    """
    for r in summands:
        seed += r
    return seed


def inline_product(factors, seed):
    """Computes a product, using the __imul__ operator.
    Args:
        seed (T): The starting total. The unit value.
        factors (iterable[T]): Values to multiply (with *=) into the total.
    Returns:
        T: The result of multiplying all the factors into the unit value.
    """
    for r in factors:
        seed *= r
    return seed


def _bravyi_kitaev_interaction_operator(interaction_operator, n_qubits):
    """Implementation of the Bravyi-Kitaev transformation for OpenFermion
    Interaction Operators. This implementation is equivalent to that described
    in arXiv:1208.5986, and has been written to optimize compute time by using
    algebraic expressions for general products a_i^dagger a_j^dagger as outlined
    in Table II of Seeley, Richard, Love.
    """

    one_body = interaction_operator.one_body_tensor
    two_body = interaction_operator.two_body_tensor
    constant_term = interaction_operator.constant

    # Compute the number of qubits.
    N = len(one_body)
    if n_qubits is None:
        n_qubits = N
    if n_qubits < N:
        raise ValueError("Invalid number of qubits specified.")

    qubit_hamiltonian = QubitOperator()
    qubit_hamiltonian_op = []
    qubit_hamiltonian_coef = []

    #  For cases A - F see Table II of Seeley, Richard, Love
    for i in range(n_qubits):
        # A. Number operators: n_i
        if abs(one_body[i, i]) > 0:
            qubit_hamiltonian += _qubit_operator_creation(
                *_seeley_richard_love(i, i, one_body[i, i], n_qubits)
            )

        for j in range(i):
            # Case B: Coulomb and exchange operators
            if abs(one_body[i, j]) > 0:
                operators, coef_list = _seeley_richard_love(
                    i, j, one_body[i, j], n_qubits
                )
                qubit_hamiltonian_op.extend(operators)
                qubit_hamiltonian_coef.extend(coef_list)

                operators, coef_list = _seeley_richard_love(
                    j, i, one_body[i, j].conj(), n_qubits
                )
                qubit_hamiltonian_op.extend(operators)
                qubit_hamiltonian_coef.extend(coef_list)

            coef = _two_body_coef(two_body, i, j, j, i) / 4
            if abs(coef) > 0:
                qubit_hamiltonian_op.append(
                    tuple((index, "Z") for index in _occupation_set(i))
                )
                qubit_hamiltonian_op.append(
                    tuple((index, "Z") for index in _occupation_set(j))
                )
                qubit_hamiltonian_op.append(
                    tuple((index, "Z") for index in _F_ij_set(i, j))
                )

                qubit_hamiltonian_coef.append(-coef)
                qubit_hamiltonian_coef.append(-coef)
                qubit_hamiltonian_coef.append(coef)
                constant_term += coef

    # C. Number-excitation operators: n_i a_j^d a_k
    for i in range(n_qubits):
        for j in range(n_qubits):
            for k in range(j):
                if i not in (j, k):
                    coef = _two_body_coef(two_body, i, j, k, i)

                    if abs(coef) > 0:
                        number = _qubit_operator_creation(
                            *_seeley_richard_love(i, i, 1, n_qubits)
                        )

                        excitation_op, excitation_coef = _seeley_richard_love(
                            j, k, coef, n_qubits
                        )
                        operators_hc, coef_list_hc = _seeley_richard_love(
                            k, j, coef.conj(), n_qubits
                        )
                        excitation_op.extend(operators_hc)
                        excitation_coef.extend(coef_list_hc)
                        excitation = _qubit_operator_creation(
                            excitation_op, excitation_coef
                        )

                        number *= excitation
                        qubit_hamiltonian += number

    # D. Double-excitation operators: c_i^d c_j^d c_k c_l
    for i in range(n_qubits):
        for j in range(i):
            for k in range(j):
                for l in range(k):
                    coef = -_two_body_coef(two_body, i, j, k, l)
                    if abs(coef) > 0:
                        qubit_hamiltonian += _hermitian_one_body_product(
                            i, j, k, l, coef, n_qubits
                        )

                    coef = -_two_body_coef(two_body, i, k, j, l)
                    if abs(coef) > 0:
                        qubit_hamiltonian += _hermitian_one_body_product(
                            i, k, j, l, coef, n_qubits
                        )

                    coef = -_two_body_coef(two_body, i, l, j, k)
                    if abs(coef) > 0:
                        qubit_hamiltonian += _hermitian_one_body_product(
                            i, l, j, k, coef, n_qubits
                        )

    qubit_hamiltonian_op.append(())
    qubit_hamiltonian_coef.append(constant_term)
    qubit_hamiltonian += _qubit_operator_creation(
        qubit_hamiltonian_op, qubit_hamiltonian_coef
    )

    return qubit_hamiltonian


def _two_body_coef(two_body, a, b, c, d):
    return (
        two_body[a, b, c, d]
        - two_body[a, b, d, c]
        + two_body[b, a, d, c]
        - two_body[b, a, c, d]
    )


def _hermitian_one_body_product(a, b, c, d, coef, n_qubits):
    """Takes the 4 indices for a two-body operator and constructs the
    Bravyi-Kitaev form by splitting the two-body into 2 one-body operators,
    multiplying them together and then re-adding the Hermitian conjugate to
    give a Hermitian operator."""

    c_dag_c_ac = _qubit_operator_creation(*_seeley_richard_love(a, c, coef, n_qubits))
    c_dag_c_bd = _qubit_operator_creation(*_seeley_richard_love(b, d, 1, n_qubits))
    c_dag_c_ac *= c_dag_c_bd
    hermitian_sum = c_dag_c_ac

    c_dag_c_ca = _qubit_operator_creation(
        *_seeley_richard_love(c, a, coef.conj(), n_qubits)
    )
    c_dag_c_db = _qubit_operator_creation(*_seeley_richard_love(d, b, 1, n_qubits))
    c_dag_c_ca *= c_dag_c_db

    hermitian_sum += c_dag_c_ca
    return hermitian_sum


def _qubit_operator_creation(operators, coefficents):
    """Takes a list of tuples for operators/indices, and another for
    coefficents"""

    qubit_operator = QubitOperator()

    for index in zip(operators, coefficents):
        qubit_operator += QubitOperator(index[0], index[1])

    return qubit_operator


def _seeley_richard_love(i, j, coef, n_qubits):
    """Algebraic expressions for general products of the form a_i^d a_j term in
    the Bravyi-Kitaev basis. These expressions vary in form depending on the
    parity of the indices i and j, as well as onthe overlaps between the parity
    and update sets of the indices"""

    seeley_richard_love_op = []
    seeley_richard_love_coef = []
    coef *= 0.25
    # Case 0
    if i == j:  # Simplifies to the number operator
        seeley_richard_love_op.append(
            tuple((index, "Z") for index in _occupation_set(i))
        )
        seeley_richard_love_coef.append(-coef * 2)

        seeley_richard_love_op.append(())
        seeley_richard_love_coef.append(coef * 2)

    # Case 1
    elif i % 2 == 0 and j % 2 == 0:
        x_pad = tuple((index, "X") for index in _U_diff_a_set(i, j, n_qubits))
        y_pad = tuple((index, "Y") for index in _alpha_set(i, j, n_qubits))
        z_pad = tuple((index, "Z") for index in _P0_ij_diff_a_set(i, j, n_qubits))

        left_pad = x_pad + y_pad + z_pad

        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "X")))
        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "Y")))
        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "X")))
        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "Y")))

        if i < j:
            seeley_richard_love_coef.append(coef)
            seeley_richard_love_coef.append(-coef)
            seeley_richard_love_coef.append(complex(0, -coef))
            seeley_richard_love_coef.append(complex(0, -coef))

        else:  # whenever i < j introduce phase of -j
            seeley_richard_love_coef.append(complex(0, -coef))
            seeley_richard_love_coef.append(complex(0, coef))
            seeley_richard_love_coef.append(-coef)
            seeley_richard_love_coef.append(-coef)

    # Case 2
    elif i % 2 == 1 and j % 2 == 0 and i not in _parity_set(j):
        x_pad = tuple((index, "X") for index in _U_diff_a_set(i, j, n_qubits))
        y_pad = tuple((index, "Y") for index in _alpha_set(i, j, n_qubits))

        left_pad = x_pad + y_pad

        right_pad_1 = tuple(
            (index, "Z") for index in _P0_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        )
        right_pad_2 = tuple(
            (index, "Z") for index in _P2_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        )

        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "X")) + right_pad_1)
        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "X")) + right_pad_1)
        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "Y")) + right_pad_2)
        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "Y")) + right_pad_2)

        if i < j:
            seeley_richard_love_coef.append(coef)
            seeley_richard_love_coef.append(complex(0, -coef))
            seeley_richard_love_coef.append(-coef)
            seeley_richard_love_coef.append(complex(0, -coef))

        else:
            seeley_richard_love_coef.append(complex(0, -coef))
            seeley_richard_love_coef.append(-coef)
            seeley_richard_love_coef.append(complex(0, coef))
            seeley_richard_love_coef.append(-coef)

    # Case 3
    elif i % 2 == 1 and j % 2 == 0 and i in _parity_set(j):
        left_pad = tuple((index, "X") for index in _U_ij_set(i, j, n_qubits))
        right_pad_1 = tuple((index, "Z") for index in _P0_ij_set(i, j) - {i})
        right_pad_2 = tuple((index, "Z") for index in _P2_ij_set(i, j) - {i})

        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "Y")) + right_pad_1)
        seeley_richard_love_coef.append(coef)

        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "Y")) + right_pad_1)
        seeley_richard_love_coef.append(complex(0, -coef))

        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "X")) + right_pad_2)
        seeley_richard_love_coef.append(coef)

        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "X")) + right_pad_2)
        seeley_richard_love_coef.append(complex(0, coef))

    # Case 4
    elif (
        i % 2 == 0
        and j % 2 == 1
        and i not in _parity_set(j)
        and j not in _update_set(i, n_qubits)
    ):
        x_pad = tuple((index, "X") for index in _U_diff_a_set(i, j, n_qubits))
        y_pad = tuple((index, "Y") for index in _alpha_set(i, j, n_qubits))
        left_pad = x_pad + y_pad

        right_pad_1 = tuple(
            (index, "Z") for index in _P0_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        )
        right_pad_2 = tuple(
            (index, "Z") for index in _P1_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        )

        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "Y")) + right_pad_1)
        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "X")) + right_pad_1)
        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "X")) + right_pad_2)
        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "Y")) + right_pad_2)

        if i < j:
            seeley_richard_love_coef.append(-coef)
            seeley_richard_love_coef.append(complex(0, -coef))
            seeley_richard_love_coef.append(coef)
            seeley_richard_love_coef.append(complex(0, -coef))
        else:
            seeley_richard_love_coef.append(complex(0, coef))
            seeley_richard_love_coef.append(-coef)
            seeley_richard_love_coef.append(complex(0, -coef))
            seeley_richard_love_coef.append(-coef)

    # Case 5
    elif (
        i % 2 == 0
        and j % 2 == 1
        and i not in _parity_set(j)
        and j in _update_set(i, n_qubits)
    ):
        x_range_1 = _U_ij_set(i, j, n_qubits) - {j}
        left_pad_1 = tuple((index, "X") for index in x_range_1)

        x_range_2 = x_range_1 - _alpha_set(i, j, n_qubits)
        left_pad_2 = tuple((index, "X") for index in x_range_2)

        y_pad = tuple((index, "Y") for index in _alpha_set(i, j, n_qubits))
        z_range_1 = _P0_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        z_pad = tuple((index, "Z") for index in z_range_1)
        right_pad_1 = y_pad + z_pad

        z_range_2 = _P1_ij_set(i, j).union({j})
        right_pad_2 = tuple((index, "Z") for index in z_range_2)

        seeley_richard_love_op.append(left_pad_2 + ((i, "Y"),) + right_pad_1)
        seeley_richard_love_coef.append(-coef)

        seeley_richard_love_op.append(left_pad_2 + ((i, "X"),) + right_pad_1)
        seeley_richard_love_coef.append(
            complex(0, -coef)
        )  # Phase flip of -1 relative to original paper

        seeley_richard_love_op.append(left_pad_1 + ((i, "Y"),) + right_pad_2)
        seeley_richard_love_coef.append(complex(0, coef))

        seeley_richard_love_op.append(left_pad_1 + ((i, "X"),) + right_pad_2)
        seeley_richard_love_coef.append(-coef)

    # Case 6
    elif (
        i % 2 == 0
        and j % 2 == 1
        and i in _parity_set(j)
        and j in _update_set(i, n_qubits)
    ):
        left_pad = tuple((index, "X") for index in _U_ij_set(i, j, n_qubits) - {j})
        right_pad = tuple((index, "Z") for index in _P1_ij_set(i, j).union({j}))

        seeley_richard_love_op.append(left_pad + ((i, "X"),))
        seeley_richard_love_coef.append(coef)

        seeley_richard_love_op.append(left_pad + ((i, "Y"),))
        seeley_richard_love_coef.append(complex(0, -coef))

        seeley_richard_love_op.append(left_pad + ((i, "Y"),) + right_pad)
        seeley_richard_love_coef.append(complex(0, coef))

        seeley_richard_love_op.append(left_pad + ((i, "X"),) + right_pad)
        seeley_richard_love_coef.append(-coef)

    # Case 7
    elif (
        i % 2 == 1
        and j % 2 == 1
        and i not in _parity_set(j)
        and j not in _update_set(i, n_qubits)
    ):
        x_pad = tuple((index, "X") for index in _U_diff_a_set(i, j, n_qubits))
        y_pad = tuple((index, "Y") for index in _alpha_set(i, j, n_qubits))
        left_pad = x_pad + y_pad

        z_range_1 = _P0_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        right_pad_1 = tuple((index, "Z") for index in z_range_1)

        z_range_2 = _P1_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        right_pad_2 = tuple((index, "Z") for index in z_range_2)

        z_range_3 = _P2_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        right_pad_3 = tuple((index, "Z") for index in z_range_3)

        z_range_4 = _P3_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        right_pad_4 = tuple((index, "Z") for index in z_range_4)

        if i < j:
            seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "X")) + right_pad_1)
            seeley_richard_love_coef.append(complex(0, -coef))

            seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "X")) + right_pad_2)
            seeley_richard_love_coef.append(coef)

            seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "Y")) + right_pad_3)
            seeley_richard_love_coef.append(-coef)

            seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "Y")) + right_pad_4)
            seeley_richard_love_coef.append(complex(0, -coef))

        else:
            seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "X")) + right_pad_1)
            seeley_richard_love_coef.append(-coef)

            seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "X")) + right_pad_2)
            seeley_richard_love_coef.append(complex(0, -coef))

            seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "Y")) + right_pad_3)
            seeley_richard_love_coef.append(complex(0, coef))

            seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "Y")) + right_pad_4)
            seeley_richard_love_coef.append(-coef)

    # Case 8
    elif (
        i % 2 == 1
        and j % 2 == 1
        and i in _parity_set(j)
        and j not in _update_set(i, n_qubits)
    ):
        left_pad = tuple((index, "X") for index in _U_ij_set(i, j, n_qubits))

        z_range_1 = _P0_ij_set(i, j) - {i}
        right_pad_1 = tuple((index, "Z") for index in z_range_1)

        z_range_2 = _P1_ij_set(i, j) - {i}
        right_pad_2 = tuple((index, "Z") for index in z_range_2)

        z_range_3 = _P2_ij_set(i, j) - {i}
        right_pad_3 = tuple((index, "Z") for index in z_range_3)

        z_range_4 = _P3_ij_set(i, j) - {i}
        right_pad_4 = tuple((index, "Z") for index in z_range_4)

        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "Y")) + right_pad_1)
        seeley_richard_love_coef.append(complex(0, -coef))

        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "Y")) + right_pad_2)
        seeley_richard_love_coef.append(coef)

        seeley_richard_love_op.append(left_pad + ((j, "X"), (i, "X")) + right_pad_3)
        seeley_richard_love_coef.append(coef)

        seeley_richard_love_op.append(left_pad + ((j, "Y"), (i, "X")) + right_pad_4)
        seeley_richard_love_coef.append(complex(0, coef))

    # Case 9
    elif (
        i % 2 == 1
        and j % 2 == 1
        and i not in _parity_set(j)
        and j in _update_set(i, n_qubits)
    ):
        x_range_1 = _U_ij_set(i, j, n_qubits) - {j}
        left_pad_3 = tuple((index, "X") for index in x_range_1)

        x_range_2 = x_range_1 - _alpha_set(i, j, n_qubits)
        left_pad_1 = tuple((index, "X") for index in x_range_2)

        x_range_3 = x_range_1.union({i}) - _alpha_set(i, j, n_qubits)
        left_pad_2 = tuple((index, "X") for index in x_range_3)

        z_range_1 = _P2_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        z_pad_1 = tuple((index, "Z") for index in z_range_1)
        y_range = _alpha_set(i, j, n_qubits)
        y_pad = tuple((index, "Y") for index in y_range)
        right_pad_1 = z_pad_1 + y_pad

        z_range_2 = _P0_ij_set(i, j) - _alpha_set(i, j, n_qubits)
        z_pad_2 = tuple((index, "Z") for index in z_range_2)
        right_pad_2 = z_pad_2 + y_pad

        z_range_3 = _P1_ij_set(i, j).union({j})
        right_pad_3 = tuple((index, "Z") for index in z_range_3)

        z_range_4 = _P3_ij_set(i, j).union({j})
        right_pad_4 = tuple((index, "Z") for index in z_range_4)

        seeley_richard_love_op.append(left_pad_1 + ((i, "Y"),) + right_pad_1)
        seeley_richard_love_coef.append(-coef)  # phase of -j relative to original paper

        seeley_richard_love_op.append(left_pad_2 + right_pad_2)
        seeley_richard_love_coef.append(
            complex(0, -coef)
        )  # phase of -j relative to original paper

        seeley_richard_love_op.append(left_pad_3 + ((i, "X"),) + right_pad_3)
        seeley_richard_love_coef.append(-coef)

        seeley_richard_love_op.append(left_pad_3 + ((i, "Y"),) + right_pad_4)
        seeley_richard_love_coef.append(complex(0, coef))

    # Case 10
    elif (
        i % 2 == 1
        and j % 2 == 1
        and i in _parity_set(j)
        and j in _update_set(i, n_qubits)
    ):
        left_pad = tuple((index, "X") for index in _U_ij_set(i, j, n_qubits) - {j})
        right_pad_1 = tuple((index, "Z") for index in _P0_ij_set(i, j) - {i})
        right_pad_2 = tuple((index, "Z") for index in _P2_ij_set(i, j) - {i})
        right_pad_3 = tuple((index, "Z") for index in _P1_ij_set(i, j))
        right_pad_4 = tuple((index, "Z") for index in _P3_ij_set(i, j))

        seeley_richard_love_op.append(left_pad + ((i, "Y"),) + right_pad_1)
        seeley_richard_love_coef.append(complex(0, -coef))

        seeley_richard_love_op.append(left_pad + ((i, "X"),) + right_pad_2)
        seeley_richard_love_coef.append(coef)

        seeley_richard_love_op.append(left_pad + ((j, "Z"), (i, "X")) + right_pad_3)
        seeley_richard_love_coef.append(-coef)

        seeley_richard_love_op.append(left_pad + ((j, "Z"), (i, "Y")) + right_pad_4)
        seeley_richard_love_coef.append(complex(0, coef))

    return seeley_richard_love_op, seeley_richard_love_coef


def _remainder_set(index):
    return _parity_set(index) - _occupation_set(index)


def _F_ij_set(i, j):
    return _occupation_set(i).symmetric_difference(_occupation_set(j))


def _P0_ij_set(i, j):
    """The symmetric difference of sets P(i) and P(j)."""
    return _parity_set(i).symmetric_difference(_parity_set(j))


def _P1_ij_set(i, j):
    """The symmetric difference of sets P(i) and R(j)."""
    return _parity_set(i).symmetric_difference(_remainder_set(j))


def _P2_ij_set(i, j):
    """The symmetric difference of sets R(i) and P(j)."""
    return _remainder_set(i).symmetric_difference(_parity_set(j))


def _P3_ij_set(i, j):
    """The symmetric difference of sets R(i) and R(j)."""
    return _remainder_set(i).symmetric_difference(_remainder_set(j))


def _U_ij_set(i, j, n_qubits):
    """The symmetric difference of sets U(i) and U(j)"""
    return _update_set(i, n_qubits).symmetric_difference(_update_set(j, n_qubits))


def _U_diff_a_set(i, j, n_qubits):
    """Calculates the set {member U_ij diff alpha_ij}."""
    return _U_ij_set(i, j, n_qubits) - _alpha_set(i, j, n_qubits)


def _P0_ij_diff_a_set(i, j, n_qubits):
    """Calculates the set {member P_ij diff alpha_ij}."""
    return _P0_ij_set(i, j) - _alpha_set(i, j, n_qubits)


def _alpha_set(i, j, n_qubits):
    return _update_set(i, n_qubits).intersection(_parity_set(j))



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Any, Callable, Dict, List

import numpy as np
import rapidjson as json

from orquestra.quantum.openfermion import (
    InteractionOperator,
    InteractionRDM,
    IsingOperator,
    QubitOperator,
    SymbolicOperator,
)
from orquestra.quantum.typing import AnyPath, LoadSource
from orquestra.quantum.utils import convert_array_to_dict, convert_dict_to_array


def convert_interaction_op_to_dict(op: InteractionOperator) -> Dict[str, Any]:
    """Convert an InteractionOperator to a dictionary.
    Args:
        op: the operator
    Returns:
        dictionary: the dictionary representation
    """

    dictionary: Dict[str, Any] = {}
    dictionary["constant"] = convert_array_to_dict(np.array(op.constant))
    dictionary["one_body_tensor"] = convert_array_to_dict(np.array(op.one_body_tensor))
    dictionary["two_body_tensor"] = convert_array_to_dict(np.array(op.two_body_tensor))

    return dictionary


def convert_dict_to_interaction_op(dictionary: dict) -> InteractionOperator:
    """Get an InteractionOperator from a dictionary.
    Args:
        dictionary: the dictionary representation
    Returns:
        op: the operator
    """

    # The tolist method is used to convert the constant from a zero-dimensional array to
    # a float/complex
    constant = convert_dict_to_array(dictionary["constant"]).tolist()

    one_body_tensor = convert_dict_to_array(dictionary["one_body_tensor"])
    two_body_tensor = convert_dict_to_array(dictionary["two_body_tensor"])

    return InteractionOperator(constant, one_body_tensor, two_body_tensor)


def load_interaction_operator(file: LoadSource) -> InteractionOperator:
    """Load an interaction operator object from a file.
    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        op: the operator.
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    return convert_dict_to_interaction_op(data)


def save_interaction_operator(
    interaction_operator: InteractionOperator, filename: AnyPath
) -> None:
    """Save an interaction operator to file.
    Args:
        interaction_operator: the operator to be saved
        filename: the name of the file
    """

    with open(filename, "w") as f:
        f.write(
            json.dumps(convert_interaction_op_to_dict(interaction_operator), indent=2)
        )


def convert_dict_to_qubitop(dictionary: dict) -> QubitOperator:
    """Get a QubitOperator from a dictionary.
    Args:
        dictionary: the dictionary representation
    Returns:
        op: the operator
    """
    return convert_dict_to_operator(dictionary, QubitOperator)


def convert_qubitop_to_dict(op: QubitOperator) -> Dict[str, Any]:
    """Convert a QubitOperator to a dictionary.
    Args:
        op: the operator
    Returns:
        dictionary: the dictionary representation
    """

    dictionary: Dict[str, Any] = {}
    dictionary["terms"] = []
    for term in op.terms:
        term_dict: Dict[str, Any] = {
            "pauli_ops": [{"qubit": op[0], "op": op[1]} for op in term]
        }

        if isinstance(op.terms[term], complex):
            term_dict["coefficient"] = {
                "real": op.terms[term].real,
                "imag": op.terms[term].imag,
            }
        else:
            term_dict["coefficient"] = {"real": op.terms[term].real}

        dictionary["terms"].append(term_dict)

    return dictionary


def convert_dict_to_operator(
    dictionary: dict, constructor: Callable
) -> SymbolicOperator:
    full_operator = constructor()
    for term_dict in dictionary["terms"]:
        operator = []
        for pauli_op in term_dict["pauli_ops"]:
            operator.append((pauli_op["qubit"], pauli_op["op"]))
        coefficient = term_dict["coefficient"]["real"]
        if term_dict["coefficient"].get("imag"):
            coefficient += 1j * term_dict["coefficient"]["imag"]
        full_operator += constructor(operator, coefficient)

    return full_operator


def save_qubit_operator(qubit_operator: QubitOperator, filename: AnyPath) -> None:
    """Save a qubit operator to file.
    Args:
        qubit_operator: the operator to be saved
        filename: the name of the file
    """

    with open(filename, "w") as f:
        f.write(json.dumps(convert_qubitop_to_dict(qubit_operator), indent=2))


def load_qubit_operator(file: LoadSource) -> QubitOperator:
    """Load an operator object from a file.
    Args:
        file: the name of the file, or a file-like object.
    Returns:
        op: the operator.
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    return convert_dict_to_qubitop(data)


def save_qubit_operator_set(
    qubit_operator_set: List[QubitOperator], filename: AnyPath
) -> None:
    """Save a set of qubit operators to a file.

    Args:
        qubit_operator_set: a list of QubitOperator to be saved
        file: the name of the file
    """
    dictionary: Dict[str, Any] = {}
    dictionary["qubit_operators"] = []
    for qubit_operator in qubit_operator_set:
        dictionary["qubit_operators"].append(convert_qubitop_to_dict(qubit_operator))
    with open(filename, "w") as f:
        f.write(json.dumps(dictionary, indent=2))


def load_qubit_operator_set(file: LoadSource) -> List[QubitOperator]:
    """Load a set of qubit operators from a file.

    Args:
        file: the name of the file, or a file-like object.

    Returns:
        qubit_operator_set: a list of QubitOperator objects
    """
    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    qubit_operator_set = []
    for qubit_operator_dict in data["qubit_operators"]:
        qubit_operator_set.append(convert_dict_to_qubitop(qubit_operator_dict))
    return qubit_operator_set


def get_pauli_strings(qubit_operator: QubitOperator) -> List[str]:
    """Convert a qubit operator into a list of Pauli strings.

    Args:
        qubit_operator: a QubitOperator to be converted

    Returns:
        pauli_strings: list of Pauli strings
    """
    pauli_strings = []
    term_list = list(qubit_operator.terms.keys())
    for term in term_list:
        pauli_list = term
        pauli_string = ""
        for pauli in pauli_list:
            pauli_string += pauli[1] + str(pauli[0])
        pauli_strings.append(pauli_string)

    return pauli_strings


def convert_isingop_to_dict(op: IsingOperator) -> dict:
    """Convert an IsingOperator to a dictionary.

    Args:
        op: the operator

    Returns:
        dictionary: the dictionary representation
    """

    dictionary = convert_qubitop_to_dict(op)
    return dictionary


def convert_dict_to_isingop(dictionary: dict) -> IsingOperator:
    """Get a IsingOperator from a dictionary.

    Args:
        dictionary: the dictionary representation

    Returns:
        op: the operator
    """
    return convert_dict_to_operator(dictionary, IsingOperator)


def load_ising_operator(file: LoadSource) -> IsingOperator:
    """Load an Ising operator object from a file.

    Args:
        file: the name of the file, or a file-like object.

    Returns:
        op: the operator.
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    return convert_dict_to_isingop(data)


def save_ising_operator(ising_operator: IsingOperator, filename: AnyPath) -> None:
    """Save an Ising operator to file.

    Args:
        op: the operator to be saved
        filename: the name of the file
    """

    with open(filename, "w") as f:
        f.write(json.dumps(convert_isingop_to_dict(ising_operator), indent=2))


def save_parameter_grid_evaluation(parameter_grid_evaluation, filename):
    """Save a list of parameter grid evaluations to file

    Args:
        parameter_grid_evaluation (list): List of dicts with a value estimate object
            under the "value" field
        file (str or file-like object): the name of the file, or a file-like object
    """
    full_dict = {}

    for evaluation in parameter_grid_evaluation:
        value = evaluation["value"].to_dict()
        evaluation["value"] = value
    full_dict["values_set"] = parameter_grid_evaluation

    with open(filename, "w") as f:
        f.write(json.dumps(full_dict, indent=2))


def convert_interaction_rdm_to_dict(op):
    """Convert an InteractionRDM to a dictionary.
    Args:
        op (openfermion.ops.InteractionRDM): the operator
    Returns:
        dictionary (dict): the dictionary representation
    """

    dictionary = {}
    dictionary["one_body_tensor"] = convert_array_to_dict(np.array(op.one_body_tensor))
    dictionary["two_body_tensor"] = convert_array_to_dict(np.array(op.two_body_tensor))

    return dictionary


def convert_dict_to_interaction_rdm(dictionary):
    """Get an InteractionRDM from a dictionary.
    Args:
        dictionary (dict): the dictionary representation
    Returns:
        op (openfermion.ops.InteractionRDM): the operator
    """

    one_body_tensor = convert_dict_to_array(dictionary["one_body_tensor"])
    two_body_tensor = convert_dict_to_array(dictionary["two_body_tensor"])

    return InteractionRDM(one_body_tensor, two_body_tensor)


def load_interaction_rdm(file: LoadSource) -> InteractionRDM:
    """Load an interaction RDM object from a file.
    Args:
        file: a file-like object to load the interaction RDM from.

    Returns:
        The interaction RDM.
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    return convert_dict_to_interaction_rdm(data)


def save_interaction_rdm(interaction_rdm: InteractionRDM, filename: AnyPath) -> None:
    """Save an interaction operator to file.
    Args:
        interaction_operator: the operator to be saved
        filename: the name of the file
    """

    with open(filename, "w") as f:
        f.write(json.dumps(convert_interaction_rdm_to_dict(interaction_rdm), indent=2))



################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
OpenFermion

For more information, examples, or tutorials visit our website:

www.openfermion.org
"""
# isort: skip_file


from orquestra.quantum.openfermion.chem import (
    MolecularData,
    angstroms_to_bohr,
    bohr_to_angstroms,
    geometry_from_file,
    load_molecular_hamiltonian,
    make_reduced_hamiltonian,
    name_molecule,
)

from orquestra.quantum.openfermion.hamiltonians import (
    fermi_hubbard,
    interaction_operator,
)

from orquestra.quantum.openfermion.linalg import (
    sparse_eigenspectrum,
    kronecker_operators,
    jw_number_restrict_operator,
    jw_number_indices,
    jw_hartree_fock_state,
    jw_configuration_state,
    jordan_wigner_sparse,
    wrapped_kronecker,
    jordan_wigner_ladder_sparse,
    eigenspectrum,
    expectation,
    get_density_matrix,
    get_ground_state,
    get_number_preserving_sparse_operator,
    get_sparse_operator,
    inner_product,
    jw_get_ground_state_at_particle_number,
    qubit_operator_sparse,
)

from orquestra.quantum.openfermion.measurements import get_interaction_rdm
from orquestra.quantum.openfermion.ops import (
    FermionOperator,
    InteractionOperator,
    InteractionOperatorError,
    InteractionRDM,
    InteractionRDMError,
    IsingOperator,
    PolynomialTensor,
    PolynomialTensorError,
    QubitOperator,
    SymbolicOperator,
    general_basis_change,
)

# Shifted here to fix circular dependencies
from orquestra.quantum.openfermion.circuits import (
    uccsd_convert_amplitude_format,
    uccsd_generator,
    uccsd_singlet_generator,
    uccsd_singlet_get_packed_amplitudes,
    uccsd_singlet_paramsize,
)

from orquestra.quantum.openfermion.transforms import (
    jordan_wigner_one_body,
    jordan_wigner_two_body,
    bravyi_kitaev,
    check_no_sympy,
    chemist_ordered,
    freeze_orbitals,
    get_fermion_operator,
    get_interaction_operator,
    get_molecular_data,
    jordan_wigner,
    normal_ordered,
    normal_ordered_ladder_term,
    prune_unused_indices,
    reorder,
    reverse_jordan_wigner,
)
from orquestra.quantum.openfermion.utils import (
    OperatorSpecificationError,
    OperatorUtilsError,
    anticommutator,
    commutator,
    count_qubits,
    double_commutator,
    down_index,
    get_file_path,
    hermitian_conjugated,
    is_hermitian,
    is_identity,
    load_operator,
    save_operator,
    up_index,
    up_then_down,
)

from orquestra.quantum.openfermion.zapata_utils import (
    # _utils.py
    get_qubitop_from_matrix,
    get_qubitop_from_coeffs_and_labels,
    generate_random_qubitop,
    evaluate_qubit_operator,
    evaluate_qubit_operator_list,
    reverse_qubit_order,
    get_expectation_value,
    change_operator_type,
    get_fermion_number_operator,
    get_diagonal_component,
    get_polynomial_tensor,
    create_circuits_from_qubit_operator,
    get_ground_state_rdm_from_qubit_op,
    remove_inactive_orbitals,
    hf_rdm,
    # _io.py
    convert_interaction_op_to_dict,
    convert_dict_to_interaction_op,
    load_interaction_operator,
    save_interaction_operator,
    convert_dict_to_qubitop,
    convert_qubitop_to_dict,
    convert_dict_to_operator,
    save_qubit_operator,
    load_qubit_operator,
    save_qubit_operator_set,
    load_qubit_operator_set,
    get_pauli_strings,
    convert_isingop_to_dict,
    convert_dict_to_isingop,
    load_ising_operator,
    save_ising_operator,
    save_parameter_grid_evaluation,
    convert_interaction_rdm_to_dict,
    convert_dict_to_interaction_rdm,
    load_interaction_rdm,
    save_interaction_rdm,
)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Module to compute commutators, with optimizations for specific systems."""

import numpy

from orquestra.quantum.openfermion.ops.operators import FermionOperator
from orquestra.quantum.openfermion.transforms.opconversions.term_reordering import (
    normal_ordered,
)


def commutator(operator_a, operator_b):
    """Compute the commutator of two operators.

    Args:
        operator_a, operator_b: Operators in commutator. Any operators
            are accepted so long as implicit subtraction and multiplication are
            supported; e.g. QubitOperators, FermionOperators, BosonOperators,
            or Scipy sparse matrices. 2D Numpy arrays are also supported.

    Raises:
        TypeError: operator_a and operator_b are not of the same type.
    """
    if type(operator_a) != type(operator_b):
        raise TypeError("operator_a and operator_b are not of the same type.")
    if isinstance(operator_a, numpy.ndarray):
        result = operator_a.dot(operator_b)
        result -= operator_b.dot(operator_a)
    else:
        result = operator_a * operator_b
        result -= operator_b * operator_a
    return result


def anticommutator(operator_a, operator_b):
    """Compute the anticommutator of two operators.

    Args:
        operator_a, operator_b: Operators in anticommutator. Any operators
            are accepted so long as implicit addition and multiplication are
            supported; e.g. QubitOperators, FermionOperators, BosonOperators,
            or Scipy sparse matrices. 2D Numpy arrays are also supported.

    Raises:
        TypeError: operator_a and operator_b are not of the same type.
    """
    if type(operator_a) != type(operator_b):
        raise TypeError("operator_a and operator_b are not of the same type.")
    if isinstance(operator_a, numpy.ndarray):
        result = operator_a.dot(operator_b)
        result += operator_b.dot(operator_a)
    else:
        result = operator_a * operator_b
        result += operator_b * operator_a
    return result


def double_commutator(
    op1,
    op2,
    op3,
    indices2=None,
    indices3=None,
    is_hopping_operator2=None,
    is_hopping_operator3=None,
):
    """Return the double commutator [op1, [op2, op3]].

    Args:
        op1, op2, op3 (FermionOperators or BosonOperators): operators for
            the commutator. All three operators must be of the same type.
        indices2, indices3 (set): The indices op2 and op3 act on.
        is_hopping_operator2 (bool): Whether op2 is a hopping operator.
        is_hopping_operator3 (bool): Whether op3 is a hopping operator.

    Returns:
        The double commutator of the given operators.
    """
    if is_hopping_operator2 and is_hopping_operator3:
        indices2 = set(indices2)
        indices3 = set(indices3)
        # Determine which indices both op2 and op3 act on.
        try:
            (intersection,) = indices2.intersection(indices3)
        except ValueError:
            return FermionOperator.zero()

        # Remove the intersection from the set of indices, since it will get
        # cancelled out in the final result.
        indices2.remove(intersection)
        indices3.remove(intersection)

        # Find the indices of the final output hopping operator.
        (index2,) = indices2
        (index3,) = indices3
        coeff2 = op2.terms[list(op2.terms)[0]]
        coeff3 = op3.terms[list(op3.terms)[0]]
        commutator23 = FermionOperator(
            ((index2, 1), (index3, 0)), coeff2 * coeff3
        ) + FermionOperator(((index3, 1), (index2, 0)), -coeff2 * coeff3)
    else:
        commutator23 = normal_ordered(commutator(op2, op3))

    return normal_ordered(commutator(op1, commutator23))



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for sparse_tools.py."""
import os
import unittest

import numpy
import scipy
from scipy.linalg import eigh, norm
from scipy.sparse import csc_matrix
from scipy.special import comb

from orquestra.quantum.openfermion.chem import MolecularData
from orquestra.quantum.openfermion.config import DATA_DIRECTORY
from orquestra.quantum.openfermion.hamiltonians import fermi_hubbard
from orquestra.quantum.openfermion.hamiltonians.special_operators import number_operator
from orquestra.quantum.openfermion.linalg.sparse_tools import (
    _iterate_basis_,
    eigenspectrum,
    expectation,
    get_ground_state,
    get_number_preserving_sparse_operator,
    get_sparse_operator,
    inner_product,
    jordan_wigner_sparse,
    jw_configuration_state,
    jw_get_ground_state_at_particle_number,
    jw_hartree_fock_state,
    jw_number_indices,
    jw_number_restrict_operator,
    qubit_operator_sparse,
    sparse_eigenspectrum,
)
from orquestra.quantum.openfermion.ops.operators import FermionOperator, QubitOperator
from orquestra.quantum.openfermion.transforms.opconversions import (
    get_fermion_operator,
    jordan_wigner,
)
from orquestra.quantum.openfermion.transforms.repconversions import (
    get_interaction_operator,
)
from orquestra.quantum.openfermion.utils.operator_utils import (
    count_qubits,
    hermitian_conjugated,
    is_hermitian,
)


class EigenSpectrumTest(unittest.TestCase):
    def setUp(self):
        self.n_qubits = 5
        self.fermion_term = FermionOperator("1^ 2^ 3 4", -3.17)
        self.fermion_operator = self.fermion_term + hermitian_conjugated(
            self.fermion_term
        )
        self.qubit_operator = jordan_wigner(self.fermion_operator)
        self.interaction_operator = get_interaction_operator(self.fermion_operator)

    def test_eigenspectrum(self):
        fermion_eigenspectrum = eigenspectrum(self.fermion_operator)
        qubit_eigenspectrum = eigenspectrum(self.qubit_operator)
        interaction_eigenspectrum = eigenspectrum(self.interaction_operator)
        for i in range(2**self.n_qubits):
            self.assertAlmostEqual(fermion_eigenspectrum[i], qubit_eigenspectrum[i])
            self.assertAlmostEqual(
                fermion_eigenspectrum[i], interaction_eigenspectrum[i]
            )


class SparseOperatorTest(unittest.TestCase):
    def test_qubit_jw_fermion_integration(self):

        # Initialize a random fermionic operator.
        fermion_operator = FermionOperator(((3, 1), (2, 1), (1, 0), (0, 0)), -4.3)
        fermion_operator += FermionOperator(((3, 1), (1, 0)), 8.17)
        fermion_operator += 3.2 * FermionOperator()

        # Map to qubits and compare matrix versions.
        qubit_operator = jordan_wigner(fermion_operator)
        qubit_sparse = get_sparse_operator(qubit_operator)
        qubit_spectrum = sparse_eigenspectrum(qubit_sparse)
        fermion_sparse = jordan_wigner_sparse(fermion_operator)
        fermion_spectrum = sparse_eigenspectrum(fermion_sparse)
        self.assertAlmostEqual(
            0.0, numpy.amax(numpy.absolute(fermion_spectrum - qubit_spectrum))
        )


class JordanWignerSparseTest(unittest.TestCase):
    def test_jw_sparse_0create(self):
        expected = csc_matrix(([1], ([1], [0])), shape=(2, 2))
        self.assertTrue(
            numpy.allclose(jordan_wigner_sparse(FermionOperator("0^")).A, expected.A)
        )

    def test_jw_sparse_1annihilate(self):
        expected = csc_matrix(([1, -1], ([0, 2], [1, 3])), shape=(4, 4))
        self.assertTrue(
            numpy.allclose(jordan_wigner_sparse(FermionOperator("1")).A, expected.A)
        )

    def test_jw_sparse_0create_2annihilate(self):
        expected = csc_matrix(
            ([-1j, 1j], ([4, 6], [1, 3])), shape=(8, 8), dtype=numpy.complex128
        )
        self.assertTrue(
            numpy.allclose(
                jordan_wigner_sparse(FermionOperator("0^ 2", -1j)).A, expected.A
            )
        )

    def test_jw_sparse_0create_3annihilate(self):
        expected = csc_matrix(
            ([-1j, 1j, 1j, -1j], ([8, 10, 12, 14], [1, 3, 5, 7])),
            shape=(16, 16),
            dtype=numpy.complex128,
        )
        self.assertTrue(
            numpy.allclose(
                jordan_wigner_sparse(FermionOperator("0^ 3", -1j)).A, expected.A
            )
        )

    def test_jw_sparse_twobody(self):
        expected = csc_matrix(([1, 1], ([6, 14], [5, 13])), shape=(16, 16))
        self.assertTrue(
            numpy.allclose(
                jordan_wigner_sparse(FermionOperator("2^ 1^ 1 3")).A, expected.A
            )
        )

    def test_qubit_operator_sparse_n_qubits_too_small(self):
        with self.assertRaises(ValueError):
            qubit_operator_sparse(QubitOperator("X3"), 1)

    def test_qubit_operator_sparse_n_qubits_not_specified(self):
        expected = csc_matrix(
            ([1, 1, 1, 1], ([1, 0, 3, 2], [0, 1, 2, 3])), shape=(4, 4)
        )
        self.assertTrue(
            numpy.allclose(qubit_operator_sparse(QubitOperator("X1")).A, expected.A)
        )


class ComputationalBasisStateTest(unittest.TestCase):
    def test_computational_basis_state(self):
        comp_basis_state = jw_configuration_state([0, 2, 5], 7)
        self.assertAlmostEqual(comp_basis_state[82], 1.0)
        self.assertAlmostEqual(sum(comp_basis_state), 1.0)


class JWHartreeFockStateTest(unittest.TestCase):
    def test_jw_hartree_fock_state(self):
        hartree_fock_state = jw_hartree_fock_state(3, 7)
        self.assertAlmostEqual(hartree_fock_state[112], 1.0)
        self.assertAlmostEqual(sum(hartree_fock_state), 1.0)


class JWNumberIndicesTest(unittest.TestCase):
    def test_jw_sparse_index(self):
        """Test the indexing scheme for selecting specific particle numbers"""
        expected = [1, 2]
        calculated_indices = jw_number_indices(1, 2)
        self.assertEqual(expected, calculated_indices)

        expected = [3]
        calculated_indices = jw_number_indices(2, 2)
        self.assertEqual(expected, calculated_indices)

    def test_jw_number_indices(self):
        n_qubits = numpy.random.randint(1, 12)
        n_particles = numpy.random.randint(n_qubits + 1)

        number_indices = jw_number_indices(n_particles, n_qubits)
        subspace_dimension = len(number_indices)

        self.assertEqual(subspace_dimension, comb(n_qubits, n_particles))

        for index in number_indices:
            binary_string = bin(index)[2:].zfill(n_qubits)
            n_ones = binary_string.count("1")
            self.assertEqual(n_ones, n_particles)


class JWNumberRestrictOperatorTest(unittest.TestCase):
    def test_jw_restrict_operator(self):
        """Test the scheme for restricting JW encoded operators to number"""
        # Make a Hamiltonian that cares mostly about number of electrons
        n_qubits = 4
        target_electrons = 2
        penalty_const = 10.0
        number_sparse = jordan_wigner_sparse(number_operator(n_qubits))
        bias_sparse = jordan_wigner_sparse(
            sum(
                [FermionOperator(((i, 1), (i, 0)), 1.0) for i in range(n_qubits)],
                FermionOperator(),
            )
        )
        hamiltonian_sparse = (
            penalty_const
            * (
                number_sparse - target_electrons * scipy.sparse.identity(2**n_qubits)
            ).dot(
                number_sparse - target_electrons * scipy.sparse.identity(2**n_qubits)
            )
            + bias_sparse
        )

        restricted_hamiltonian = jw_number_restrict_operator(
            hamiltonian_sparse, target_electrons, n_qubits
        )
        true_eigvals, _ = eigh(hamiltonian_sparse.A)
        test_eigvals, _ = eigh(restricted_hamiltonian.A)

        self.assertAlmostEqual(norm(true_eigvals[:6] - test_eigvals[:6]), 0.0)

    def test_jw_restrict_operator_hopping_to_1_particle(self):
        hop = FermionOperator("3^ 1") + FermionOperator("1^ 3")
        hop_sparse = jordan_wigner_sparse(hop, n_qubits=4)
        hop_restrict = jw_number_restrict_operator(hop_sparse, 1, n_qubits=4)
        expected = csc_matrix(([1, 1], ([0, 2], [2, 0])), shape=(4, 4))

        self.assertTrue(numpy.allclose(hop_restrict.A, expected.A))

    def test_jw_restrict_operator_interaction_to_1_particle(self):
        interaction = FermionOperator("3^ 2^ 4 1")
        interaction_sparse = jordan_wigner_sparse(interaction, n_qubits=6)
        interaction_restrict = jw_number_restrict_operator(
            interaction_sparse, 1, n_qubits=6
        )
        expected = csc_matrix(([], ([], [])), shape=(6, 6))

        self.assertTrue(numpy.allclose(interaction_restrict.A, expected.A))

    def test_jw_restrict_operator_interaction_to_2_particles(self):
        interaction = FermionOperator("3^ 2^ 4 1") + FermionOperator("4^ 1^ 3 2")
        interaction_sparse = jordan_wigner_sparse(interaction, n_qubits=6)
        interaction_restrict = jw_number_restrict_operator(
            interaction_sparse, 2, n_qubits=6
        )

        dim = 6 * 5 // 2  # shape of new sparse array

        # 3^ 2^ 4 1 maps 2**4 + 2 = 18 to 2**3 + 2**2 = 12 and vice versa;
        # in the 2-particle subspace (1, 4) and (2, 3) are 7th and 9th.
        expected = csc_matrix(([-1, -1], ([7, 9], [9, 7])), shape=(dim, dim))

        self.assertTrue(numpy.allclose(interaction_restrict.A, expected.A))

    def test_jw_restrict_operator_hopping_to_1_particle_default_nqubits(self):
        interaction = FermionOperator("3^ 2^ 4 1") + FermionOperator("4^ 1^ 3 2")
        interaction_sparse = jordan_wigner_sparse(interaction, n_qubits=6)
        # n_qubits should default to 6
        interaction_restrict = jw_number_restrict_operator(interaction_sparse, 2)

        dim = 6 * 5 // 2  # shape of new sparse array

        # 3^ 2^ 4 1 maps 2**4 + 2 = 18 to 2**3 + 2**2 = 12 and vice versa;
        # in the 2-particle subspace (1, 4) and (2, 3) are 7th and 9th.
        expected = csc_matrix(([-1, -1], ([7, 9], [9, 7])), shape=(dim, dim))

        self.assertTrue(numpy.allclose(interaction_restrict.A, expected.A))


class JWGetGroundStatesByParticleNumberTest(unittest.TestCase):
    def test_jw_get_ground_state_at_particle_number_herm_conserving(self):
        # Initialize a particle-number-conserving Hermitian operator
        ferm_op = (
            FermionOperator("0^ 1")
            + FermionOperator("1^ 0")
            + FermionOperator("1^ 2")
            + FermionOperator("2^ 1")
            + FermionOperator("1^ 3", -0.4)
            + FermionOperator("3^ 1", -0.4)
        )
        jw_hamiltonian = jordan_wigner(ferm_op)
        sparse_operator = get_sparse_operator(jw_hamiltonian)
        n_qubits = 4

        num_op = get_sparse_operator(number_operator(n_qubits))

        # Test each possible particle number
        for particle_number in range(n_qubits):
            # Get the ground energy and ground state at this particle number
            energy, state = jw_get_ground_state_at_particle_number(
                sparse_operator, particle_number
            )

            # Check that it's an eigenvector with the correct eigenvalue
            self.assertTrue(numpy.allclose(sparse_operator.dot(state), energy * state))

            # Check that it has the correct particle number
            num = expectation(num_op, state)
            self.assertAlmostEqual(num, particle_number)

    def test_jw_get_ground_state_at_particle_number_hubbard(self):

        model = fermi_hubbard(2, 2, 1.0, 4.0)
        sparse_operator = get_sparse_operator(model)
        n_qubits = count_qubits(model)
        num_op = get_sparse_operator(number_operator(n_qubits))

        # Test each possible particle number
        for particle_number in range(n_qubits):
            # Get the ground energy and ground state at this particle number
            energy, state = jw_get_ground_state_at_particle_number(
                sparse_operator, particle_number
            )

            # Check that it's an eigenvector with the correct eigenvalue
            self.assertTrue(numpy.allclose(sparse_operator.dot(state), energy * state))

            # Check that it has the correct particle number
            num = expectation(num_op, state)
            self.assertAlmostEqual(num, particle_number)


class GroundStateTest(unittest.TestCase):
    def test_get_ground_state_hermitian(self):
        ground = get_ground_state(
            get_sparse_operator(QubitOperator("Y0 X1") + QubitOperator("Z0 Z1"))
        )
        expected_state = csc_matrix(
            ([1j, 1], ([1, 2], [0, 0])), shape=(4, 1), dtype=numpy.complex128
        ).A
        expected_state /= numpy.sqrt(2.0)

        self.assertAlmostEqual(ground[0], -2)
        self.assertAlmostEqual(
            numpy.absolute(expected_state.T.conj().dot(ground[1]))[0], 1.0
        )


class ExpectationTest(unittest.TestCase):
    def test_expectation_correct_sparse_matrix(self):
        operator = get_sparse_operator(QubitOperator("X0"), n_qubits=2)
        vector = numpy.array([0.0, 1.0j, 0.0, 1.0j])
        self.assertAlmostEqual(expectation(operator, vector), 2.0)

        density_matrix = scipy.sparse.csc_matrix(
            numpy.outer(vector, numpy.conjugate(vector))
        )
        self.assertAlmostEqual(expectation(operator, density_matrix), 2.0)

    def test_expectation_handles_column_vector(self):
        operator = get_sparse_operator(QubitOperator("X0"), n_qubits=2)
        vector = numpy.array([[0.0], [1.0j], [0.0], [1.0j]])
        self.assertAlmostEqual(expectation(operator, vector), 2.0)

    def test_expectation_correct_zero(self):
        operator = get_sparse_operator(QubitOperator("X0"), n_qubits=2)
        vector = numpy.array([1j, -1j, -1j, -1j])
        self.assertAlmostEqual(expectation(operator, vector), 0.0)


class InnerProductTest(unittest.TestCase):
    def test_inner_product(self):
        state_1 = numpy.array([1.0, 1.0j])
        state_2 = numpy.array([1.0, -1.0j])

        self.assertAlmostEqual(inner_product(state_1, state_1), 2.0)
        self.assertAlmostEqual(inner_product(state_1, state_2), 0.0)


class GetNumberPreservingSparseOperatorIntegrationTestLiH(unittest.TestCase):
    def setUp(self):
        # Set up molecule.
        geometry = [("Li", (0.0, 0.0, 0.0)), ("H", (0.0, 0.0, 1.45))]
        basis = "sto-3g"
        multiplicity = 1
        filename = os.path.join(DATA_DIRECTORY, "H1-Li1_sto-3g_singlet_1.45")
        self.molecule = MolecularData(geometry, basis, multiplicity, filename=filename)
        self.molecule.load()

        # Get molecular Hamiltonian.
        self.molecular_hamiltonian = self.molecule.get_molecular_hamiltonian()

        self.hubbard_hamiltonian = fermi_hubbard(
            2, 2, 1.0, 4.0, chemical_potential=0.2, magnetic_field=0.0, spinless=False
        )

    def test_exceptions(self):
        op = FermionOperator("1")
        with self.assertRaises(ValueError):
            _ = get_number_preserving_sparse_operator(op, 2, 1)

    def test_number_on_reference(self):
        sum_n_op = FermionOperator()
        sum_sparse_n_op = get_number_preserving_sparse_operator(
            sum_n_op,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False,
        )

        space_size = sum_sparse_n_op.shape[0]
        reference = numpy.zeros((space_size))
        reference[0] = 1.0

        for i in range(self.molecule.n_qubits):
            n_op = FermionOperator(((i, 1), (i, 0)))
            sum_n_op += n_op

            sparse_n_op = get_number_preserving_sparse_operator(
                n_op,
                self.molecule.n_qubits,
                self.molecule.n_electrons,
                spin_preserving=False,
            )

            sum_sparse_n_op += sparse_n_op

            expectation = reference.dot(sparse_n_op.dot(reference))

            if i < self.molecule.n_electrons:
                assert expectation == 1.0
            else:
                assert expectation == 0.0

        convert_after_adding = get_number_preserving_sparse_operator(
            sum_n_op,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False,
        )

        assert scipy.sparse.linalg.norm(convert_after_adding - sum_sparse_n_op) < 1e-9

        assert (
            reference.dot(sum_sparse_n_op.dot(reference)) - self.molecule.n_electrons
            < 1e-9
        )

    def test_space_size_correct(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
        )

        space_size = sparse_ham.shape[0]

        # Naive Hilbert space size is 2**12, or 4096.
        assert space_size == 225

    def test_hf_energy(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
        )

        space_size = sparse_ham.shape[0]
        reference = numpy.zeros((space_size))
        reference[0] = 1.0

        sparse_hf_energy = reference.dot(sparse_ham.dot(reference))

        assert numpy.linalg.norm(sparse_hf_energy - self.molecule.hf_energy) < 1e-9

    def test_one_body_hf_energy(self):
        one_body_part = self.molecular_hamiltonian
        one_body_part.two_body_tensor = numpy.zeros_like(one_body_part.two_body_tensor)

        one_body_fop = get_fermion_operator(one_body_part)
        one_body_regular_sparse_op = get_sparse_operator(one_body_fop)

        make_hf_fop = FermionOperator(((3, 1), (2, 1), (1, 1), (0, 1)))
        make_hf_sparse_op = get_sparse_operator(make_hf_fop, n_qubits=12)

        hf_state = numpy.zeros((2**12))
        hf_state[0] = 1.0
        hf_state = make_hf_sparse_op.dot(hf_state)

        regular_sparse_hf_energy = (
            hf_state.dot(one_body_regular_sparse_op.dot(hf_state))
        ).real

        one_body_sparse_op = get_number_preserving_sparse_operator(
            one_body_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
        )

        space_size = one_body_sparse_op.shape[0]
        reference = numpy.zeros((space_size))
        reference[0] = 1.0

        sparse_hf_energy = reference.dot(one_body_sparse_op.dot(reference))

        assert numpy.linalg.norm(sparse_hf_energy - regular_sparse_hf_energy) < 1e-9

    def test_ground_state_energy(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
        )

        eig_val, _ = scipy.sparse.linalg.eigsh(sparse_ham, k=1, which="SA")

        assert numpy.abs(eig_val[0] - self.molecule.fci_energy) < 1e-9

    def test_doubles_are_subset(self):
        reference_determinants = [
            [
                True,
                True,
                True,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            [
                True,
                True,
                False,
                False,
                False,
                False,
                True,
                True,
                False,
                False,
                False,
                False,
            ],
        ]

        for reference_determinant in reference_determinants:
            reference_determinant = numpy.asarray(reference_determinant)
            doubles_state_array = numpy.asarray(
                list(
                    _iterate_basis_(
                        reference_determinant, excitation_level=2, spin_preserving=True
                    )
                )
            )
            doubles_int_state_array = doubles_state_array.dot(
                1 << numpy.arange(doubles_state_array.shape[1])[::-1]
            )

            all_state_array = numpy.asarray(
                list(
                    _iterate_basis_(
                        reference_determinant, excitation_level=4, spin_preserving=True
                    )
                )
            )
            all_int_state_array = all_state_array.dot(
                1 << numpy.arange(all_state_array.shape[1])[::-1]
            )

            for item in doubles_int_state_array:
                assert item in all_int_state_array

        for reference_determinant in reference_determinants:
            reference_determinant = numpy.asarray(reference_determinant)
            doubles_state_array = numpy.asarray(
                list(
                    _iterate_basis_(
                        reference_determinant, excitation_level=2, spin_preserving=True
                    )
                )
            )
            doubles_int_state_array = doubles_state_array.dot(
                1 << numpy.arange(doubles_state_array.shape[1])[::-1]
            )

            all_state_array = numpy.asarray(
                list(
                    _iterate_basis_(
                        reference_determinant, excitation_level=4, spin_preserving=False
                    )
                )
            )
            all_int_state_array = all_state_array.dot(
                1 << numpy.arange(all_state_array.shape[1])[::-1]
            )

            for item in doubles_int_state_array:
                assert item in all_int_state_array

        for reference_determinant in reference_determinants:
            reference_determinant = numpy.asarray(reference_determinant)
            doubles_state_array = numpy.asarray(
                list(
                    _iterate_basis_(
                        reference_determinant, excitation_level=2, spin_preserving=False
                    )
                )
            )
            doubles_int_state_array = doubles_state_array.dot(
                1 << numpy.arange(doubles_state_array.shape[1])[::-1]
            )

            all_state_array = numpy.asarray(
                list(
                    _iterate_basis_(
                        reference_determinant, excitation_level=4, spin_preserving=False
                    )
                )
            )
            all_int_state_array = all_state_array.dot(
                1 << numpy.arange(all_state_array.shape[1])[::-1]
            )

            for item in doubles_int_state_array:
                assert item in all_int_state_array

    def test_full_ham_hermitian(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
        )

        assert scipy.sparse.linalg.norm(sparse_ham - sparse_ham.getH()) < 1e-9

    def test_full_ham_hermitian_non_spin_preserving(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False,
        )

        assert scipy.sparse.linalg.norm(sparse_ham - sparse_ham.getH()) < 1e-9

    def test_singles_simple_one_body_term_hermitian(self):
        fop = FermionOperator(((3, 1), (1, 0)))
        fop_conj = FermionOperator(((1, 1), (3, 0)))

        sparse_op = get_number_preserving_sparse_operator(
            fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1,
        )

        sparse_op_conj = get_number_preserving_sparse_operator(
            fop_conj,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1,
        )

        assert scipy.sparse.linalg.norm(sparse_op - sparse_op_conj.getH()) < 1e-9

    def test_singles_simple_two_body_term_hermitian(self):
        fop = FermionOperator(((3, 1), (8, 1), (1, 0), (4, 0)))
        fop_conj = FermionOperator(((4, 1), (1, 1), (8, 0), (3, 0)))

        sparse_op = get_number_preserving_sparse_operator(
            fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1,
        )

        sparse_op_conj = get_number_preserving_sparse_operator(
            fop_conj,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1,
        )

        assert scipy.sparse.linalg.norm(sparse_op - sparse_op_conj.getH()) < 1e-9

    def test_singles_repeating_two_body_term_hermitian(self):
        fop = FermionOperator(((3, 1), (1, 1), (5, 0), (1, 0)))
        fop_conj = FermionOperator(((5, 1), (1, 1), (3, 0), (1, 0)))

        sparse_op = get_number_preserving_sparse_operator(
            fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1,
        )

        sparse_op_conj = get_number_preserving_sparse_operator(
            fop_conj,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1,
        )

        assert scipy.sparse.linalg.norm(sparse_op - sparse_op_conj.getH()) < 1e-9

    def test_singles_ham_hermitian(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1,
        )

        assert scipy.sparse.linalg.norm(sparse_ham - sparse_ham.getH()) < 1e-9

    def test_singles_ham_hermitian_non_spin_preserving(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False,
            excitation_level=1,
        )

        assert scipy.sparse.linalg.norm(sparse_ham - sparse_ham.getH()) < 1e-9

    def test_cisd_energy(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=2,
        )

        eig_val, _ = scipy.sparse.linalg.eigsh(sparse_ham, k=1, which="SA")

        assert numpy.abs(eig_val[0] - self.molecule.cisd_energy) < 1e-9

    def test_cisd_energy_non_spin_preserving(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False,
            excitation_level=2,
        )

        eig_val, _ = scipy.sparse.linalg.eigsh(sparse_ham, k=1, which="SA")

        assert numpy.abs(eig_val[0] - self.molecule.cisd_energy) < 1e-9

    def test_cisd_matches_fci_energy_two_electron_hubbard(self):
        hamiltonian_fop = self.hubbard_hamiltonian

        sparse_ham_cisd = get_number_preserving_sparse_operator(
            hamiltonian_fop, 8, 2, spin_preserving=True, excitation_level=2
        )

        sparse_ham_fci = get_sparse_operator(hamiltonian_fop, n_qubits=8)

        eig_val_cisd, _ = scipy.sparse.linalg.eigsh(sparse_ham_cisd, k=1, which="SA")
        eig_val_fci, _ = scipy.sparse.linalg.eigsh(sparse_ham_fci, k=1, which="SA")

        assert numpy.abs(eig_val_cisd[0] - eig_val_fci[0]) < 1e-9

    def test_weird_determinant_matches_fci_energy_two_electron_hubbard(self):
        hamiltonian_fop = self.hubbard_hamiltonian

        sparse_ham_cisd = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            8,
            2,
            spin_preserving=True,
            excitation_level=2,
            reference_determinant=numpy.asarray(
                [False, False, True, True, False, False, False, False]
            ),
        )

        sparse_ham_fci = get_sparse_operator(hamiltonian_fop, n_qubits=8)

        eig_val_cisd, _ = scipy.sparse.linalg.eigsh(sparse_ham_cisd, k=1, which="SA")
        eig_val_fci, _ = scipy.sparse.linalg.eigsh(sparse_ham_fci, k=1, which="SA")

        assert numpy.abs(eig_val_cisd[0] - eig_val_fci[0]) < 1e-9

    def test_number_restricted_spectra_match_molecule(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham_number_preserving = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False,
        )

        sparse_ham = get_sparse_operator(hamiltonian_fop, self.molecule.n_qubits)

        sparse_ham_restricted_number_preserving = jw_number_restrict_operator(
            sparse_ham,
            n_electrons=self.molecule.n_electrons,
            n_qubits=self.molecule.n_qubits,
        )

        spectrum_from_new_sparse_method = sparse_eigenspectrum(
            sparse_ham_number_preserving
        )

        spectrum_from_old_sparse_method = sparse_eigenspectrum(
            sparse_ham_restricted_number_preserving
        )

        spectral_deviation = numpy.amax(
            numpy.absolute(
                spectrum_from_new_sparse_method - spectrum_from_old_sparse_method
            )
        )
        self.assertAlmostEqual(spectral_deviation, 0.0)

    def test_number_restricted_spectra_match_hubbard(self):
        hamiltonian_fop = self.hubbard_hamiltonian

        sparse_ham_number_preserving = get_number_preserving_sparse_operator(
            hamiltonian_fop, 8, 4, spin_preserving=False
        )

        sparse_ham = get_sparse_operator(hamiltonian_fop, 8)

        sparse_ham_restricted_number_preserving = jw_number_restrict_operator(
            sparse_ham, n_electrons=4, n_qubits=8
        )

        spectrum_from_new_sparse_method = sparse_eigenspectrum(
            sparse_ham_number_preserving
        )

        spectrum_from_old_sparse_method = sparse_eigenspectrum(
            sparse_ham_restricted_number_preserving
        )

        spectral_deviation = numpy.amax(
            numpy.absolute(
                spectrum_from_new_sparse_method - spectrum_from_old_sparse_method
            )
        )
        self.assertAlmostEqual(spectral_deviation, 0.0)


class GetSparseOperatorQubitTest(unittest.TestCase):
    def test_sparse_matrix_Y(self):
        term = QubitOperator(((0, "Y"),))
        sparse_operator = get_sparse_operator(term)
        self.assertEqual(list(sparse_operator.data), [1j, -1j])
        self.assertEqual(list(sparse_operator.indices), [1, 0])
        self.assertTrue(is_hermitian(sparse_operator))

    def test_sparse_matrix_ZX(self):
        coefficient = 2.0
        operators = ((0, "Z"), (1, "X"))
        term = QubitOperator(operators, coefficient)
        sparse_operator = get_sparse_operator(term)
        self.assertEqual(list(sparse_operator.data), [2.0, 2.0, -2.0, -2.0])
        self.assertEqual(list(sparse_operator.indices), [1, 0, 3, 2])
        self.assertTrue(is_hermitian(sparse_operator))

    def test_sparse_matrix_ZIZ(self):
        operators = ((0, "Z"), (2, "Z"))
        term = QubitOperator(operators)
        sparse_operator = get_sparse_operator(term)
        self.assertEqual(list(sparse_operator.data), [1, -1, 1, -1, -1, 1, -1, 1])
        self.assertEqual(list(sparse_operator.indices), list(range(8)))
        self.assertTrue(is_hermitian(sparse_operator))

    def test_sparse_matrix_combo(self):
        qop = QubitOperator(((0, "Y"), (1, "X")), -0.1j) + QubitOperator(
            ((0, "X"), (1, "Z")), 3.0 + 2.0j
        )
        sparse_operator = get_sparse_operator(qop)

        self.assertEqual(
            list(sparse_operator.data),
            [3 + 2j, 0.1, 0.1, -3 - 2j, 3 + 2j, -0.1, -0.1, -3 - 2j],
        )
        self.assertEqual(list(sparse_operator.indices), [2, 3, 2, 3, 0, 1, 0, 1])

    def test_sparse_matrix_zero_1qubit(self):
        sparse_operator = get_sparse_operator(QubitOperator((), 0.0), 1)
        sparse_operator.eliminate_zeros()
        self.assertEqual(len(list(sparse_operator.data)), 0)
        self.assertEqual(sparse_operator.shape, (2, 2))

    def test_sparse_matrix_zero_5qubit(self):
        sparse_operator = get_sparse_operator(QubitOperator((), 0.0), 5)
        sparse_operator.eliminate_zeros()
        self.assertEqual(len(list(sparse_operator.data)), 0)
        self.assertEqual(sparse_operator.shape, (32, 32))

    def test_sparse_matrix_identity_1qubit(self):
        sparse_operator = get_sparse_operator(QubitOperator(()), 1)
        self.assertEqual(list(sparse_operator.data), [1] * 2)
        self.assertEqual(sparse_operator.shape, (2, 2))

    def test_sparse_matrix_identity_5qubit(self):
        sparse_operator = get_sparse_operator(QubitOperator(()), 5)
        self.assertEqual(list(sparse_operator.data), [1] * 32)
        self.assertEqual(sparse_operator.shape, (32, 32))

    def test_sparse_matrix_linearity(self):
        identity = QubitOperator(())
        zzzz = QubitOperator(tuple((i, "Z") for i in range(4)), 1.0)

        sparse1 = get_sparse_operator(identity + zzzz)
        sparse2 = get_sparse_operator(identity, 4) + get_sparse_operator(zzzz)

        self.assertEqual(list(sparse1.data), [2] * 8)
        self.assertEqual(list(sparse1.indices), [0, 3, 5, 6, 9, 10, 12, 15])
        self.assertEqual(list(sparse2.data), [2] * 8)
        self.assertEqual(list(sparse2.indices), [0, 3, 5, 6, 9, 10, 12, 15])


class GetSparseOperatorFermionTest(unittest.TestCase):
    def test_sparse_matrix_zero_n_qubit(self):
        sparse_operator = get_sparse_operator(FermionOperator.zero(), 4)
        sparse_operator.eliminate_zeros()
        self.assertEqual(len(list(sparse_operator.data)), 0)
        self.assertEqual(sparse_operator.shape, (16, 16))



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import unittest

from orquestra.quantum.openfermion.ops.operators.ising_operator import IsingOperator
from orquestra.quantum.openfermion.testing.testing_utils import EqualsTester


def test_properties():
    operator = IsingOperator()
    assert operator.actions == ("Z",)
    assert operator.action_strings == ("Z",)
    assert operator.action_before_index
    assert operator.different_indices_commute


class GeneralTest(unittest.TestCase):
    """General tests."""

    def test_ising_operator(self):
        equals_tester = EqualsTester(self)

        group = [IsingOperator("Z0 Z3"), IsingOperator([(0, "Z"), (3, "Z")])]

        equals_tester.add_equality_group(*group)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sympy

from orquestra.quantum.openfermion.ops.operators import FermionOperator
from orquestra.quantum.openfermion.ops.representations import PolynomialTensor


def check_no_sympy(operator):
    """Checks whether a SymbolicOperator contains any
    sympy expressions, which will prevent it being converted
    to a PolynomialTensor

    Args:
        operator(SymbolicOperator): the operator to be tested
    """
    for key in operator.terms:
        if isinstance(operator.terms[key], sympy.Expr):
            raise TypeError(
                "This conversion is currently not supported "
                + "for operators with sympy expressions "
                + "as coefficients"
            )


def get_fermion_operator(operator):
    """Convert to FermionOperator.

    Returns:
        fermion_operator: An instance of the FermionOperator class.
    """
    if isinstance(operator, PolynomialTensor):
        return _polynomial_tensor_to_fermion_operator(operator)
    else:
        raise TypeError(
            "{} cannot be converted to FermionOperator".format(type(operator))
        )


def _polynomial_tensor_to_fermion_operator(operator):
    fermion_operator = FermionOperator()
    for term in operator:
        fermion_operator += FermionOperator(term, operator[term])
    return fermion_operator



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import itertools

import numpy

from orquestra.quantum.openfermion.ops.operators import FermionOperator
from orquestra.quantum.openfermion.ops.representations import InteractionRDM
from orquestra.quantum.openfermion.transforms.opconversions.conversions import (
    check_no_sympy,
)
from orquestra.quantum.openfermion.utils.operator_utils import count_qubits


def get_interaction_rdm(qubit_operator, n_qubits=None):
    """Build an InteractionRDM from measured qubit operators.

    Returns: An InteractionRDM object.
    """

    check_no_sympy(qubit_operator)

    # Avoid circular import.
    from orquestra.quantum.openfermion.transforms import jordan_wigner

    if n_qubits is None:
        n_qubits = count_qubits(qubit_operator)
    one_rdm = numpy.zeros((n_qubits,) * 2, dtype=complex)
    two_rdm = numpy.zeros((n_qubits,) * 4, dtype=complex)

    # One-RDM.
    for i, j in itertools.product(range(n_qubits), repeat=2):
        transformed_operator = jordan_wigner(FermionOperator(((i, 1), (j, 0))))
        for term, coefficient in transformed_operator.terms.items():
            if term in qubit_operator.terms:
                one_rdm[i, j] += coefficient * qubit_operator.terms[term]

    # Two-RDM.
    for i, j, k, l in itertools.product(range(n_qubits), repeat=4):
        transformed_operator = jordan_wigner(
            FermionOperator(((i, 1), (j, 1), (k, 0), (l, 0)))
        )
        for term, coefficient in transformed_operator.terms.items():
            if term in qubit_operator.terms:
                two_rdm[i, j, k, l] += coefficient * qubit_operator.terms[term]

    return InteractionRDM(one_rdm, two_rdm)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from .sparse_tools import (
    eigenspectrum,
    expectation,
    get_density_matrix,
    get_ground_state,
    get_number_preserving_sparse_operator,
    get_sparse_operator,
    inner_product,
    jordan_wigner_ladder_sparse,
    jordan_wigner_sparse,
    jw_configuration_state,
    jw_get_ground_state_at_particle_number,
    jw_hartree_fock_state,
    jw_number_indices,
    jw_number_restrict_operator,
    kronecker_operators,
    qubit_operator_sparse,
    sparse_eigenspectrum,
    wrapped_kronecker,
)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for qubit_operator.py."""

import warnings

import numpy
import pytest

from orquestra.quantum.openfermion.ops.operators.qubit_operator import (
    _PAULI_OPERATOR_PRODUCTS,
    QubitOperator,
)


def test_pauli_operator_product():
    correct = {
        ("I", "I"): (1.0, "I"),
        ("I", "X"): (1.0, "X"),
        ("X", "I"): (1.0, "X"),
        ("I", "Y"): (1.0, "Y"),
        ("Y", "I"): (1.0, "Y"),
        ("I", "Z"): (1.0, "Z"),
        ("Z", "I"): (1.0, "Z"),
        ("X", "X"): (1.0, "I"),
        ("Y", "Y"): (1.0, "I"),
        ("Z", "Z"): (1.0, "I"),
        ("X", "Y"): (1.0j, "Z"),
        ("X", "Z"): (-1.0j, "Y"),
        ("Y", "X"): (-1.0j, "Z"),
        ("Y", "Z"): (1.0j, "X"),
        ("Z", "X"): (1.0j, "Y"),
        ("Z", "Y"): (-1.0j, "X"),
    }
    assert _PAULI_OPERATOR_PRODUCTS == correct


def test_init_simplify():
    assert QubitOperator("X0 X0") == QubitOperator.identity()
    assert QubitOperator("X1 X1") == QubitOperator.identity()
    assert QubitOperator("Y0 Y0") == QubitOperator.identity()
    assert QubitOperator("Z0 Z0") == QubitOperator.identity()
    assert QubitOperator("X0 Y0") == QubitOperator("Z0", coefficient=1j)
    assert QubitOperator("Y0 X0") == QubitOperator("Z0", coefficient=-1j)
    assert QubitOperator("X0 Y0 Z0") == 1j * QubitOperator.identity()
    assert QubitOperator("Y0 Z0 X0") == 1j * QubitOperator.identity()
    assert QubitOperator("Z0 Y0 X0") == -1j * QubitOperator.identity()
    assert QubitOperator("X1 Y0 X1") == QubitOperator("Y0")
    assert QubitOperator("Y1 Y1 Y1") == QubitOperator("Y1")
    assert QubitOperator("Y2 Y2 Y2 Y2") == QubitOperator.identity()
    assert QubitOperator("Y3 Y3 Y3 Y3 Y3") == QubitOperator("Y3")
    assert QubitOperator("Y4 Y4 Y4 Y4 Y4 Y4") == QubitOperator.identity()
    assert QubitOperator("X0 Y1 Y0 X1") == QubitOperator("Z0 Z1")
    assert QubitOperator("X0 Y1 Z3 X2 Z3 Y0") == QubitOperator(
        "Z0 Y1 X2", coefficient=1j
    )


def test_imul_inplace():
    qubit_op = QubitOperator("X1")
    prev_id = id(qubit_op)
    qubit_op *= 3.0
    assert id(qubit_op) == prev_id


def test_different_indices_commute():
    qubit_op = QubitOperator("X1")
    assert qubit_op.different_indices_commute is True


@pytest.mark.parametrize(
    "multiplier", [0.5, 0.6j, numpy.float64(2.303), numpy.complex128(-1j)]
)
def test_imul_scalar(multiplier):
    loc_op = ((1, "X"), (2, "Y"))
    qubit_op = QubitOperator(loc_op)
    qubit_op *= multiplier
    assert qubit_op.terms[loc_op] == pytest.approx(multiplier)


def test_imul_qubit_op():
    op1 = QubitOperator(((0, "Y"), (3, "X"), (8, "Z"), (11, "X")), 3.0j)
    op2 = QubitOperator(((1, "X"), (3, "Y"), (8, "Z")), 0.5)
    op1 *= op2
    correct_term = ((0, "Y"), (1, "X"), (3, "Z"), (11, "X"))
    assert len(op1.terms) == 1
    assert correct_term in op1.terms


def test_imul_qubit_op_2():
    op3 = QubitOperator(((1, "Y"), (0, "X")), -1j)
    op4 = QubitOperator(((1, "Y"), (0, "X"), (2, "Z")), -1.5)
    op3 *= op4
    op4 *= op3
    assert ((2, "Z"),) in op3.terms
    assert op3.terms[((2, "Z"),)] == 1.5j
    assert op4.terms[((0, "X"), (1, "Y"))] == -2.25j


def test_imul_bidir():
    op_a = QubitOperator(((1, "Y"), (0, "X")), -1j)
    op_b = QubitOperator(((1, "Y"), (0, "X"), (2, "Z")), -1.5)
    op_a *= op_b
    op_b *= op_a
    assert ((2, "Z"),) in op_a.terms
    assert op_a.terms[((2, "Z"),)] == 1.5j
    assert ((0, "X"), (1, "Y")) in op_b.terms
    assert op_b.terms[((0, "X"), (1, "Y"))] == -2.25j


def test_imul_bad_multiplier():
    operator = QubitOperator(((1, "Y"), (0, "X")), -1j)
    with pytest.raises(TypeError):
        operator *= "1"


def test_mul_by_scalarzero():
    operator = QubitOperator(((1, "Y"), (0, "X")), -1j) * 0
    assert ((0, "X"), (1, "Y")) in operator.terms
    assert operator.terms[((0, "X"), (1, "Y"))] == pytest.approx(0.0)


def test_mul_bad_multiplier():
    operator = QubitOperator(((1, "Y"), (0, "X")), -1j)
    with pytest.raises(TypeError):
        operator = operator * "0.5"


def test_mul_out_of_place():
    op1 = QubitOperator(((0, "Y"), (3, "X"), (8, "Z"), (11, "X")), 3.0j)
    op2 = QubitOperator(((1, "X"), (3, "Y"), (8, "Z")), 0.5)
    op3 = op1 * op2
    correct_coefficient = 1.0j * 3.0j * 0.5
    correct_term = ((0, "Y"), (1, "X"), (3, "Z"), (11, "X"))
    assert op1 == QubitOperator(((0, "Y"), (3, "X"), (8, "Z"), (11, "X")), 3.0j)
    assert op2 == QubitOperator(((1, "X"), (3, "Y"), (8, "Z")), 0.5)
    assert op3 == QubitOperator(correct_term, correct_coefficient)


def test_mul_npfloat64():
    operator = QubitOperator(((1, "X"), (3, "Y")), 0.5)
    res = operator * numpy.float64(0.5)
    assert res == QubitOperator(((1, "X"), (3, "Y")), 0.5 * 0.5)


def test_mul_multiple_terms():
    operator = QubitOperator(((1, "X"), (3, "Y"), (8, "Z")), 0.5)
    operator += QubitOperator(((1, "Z"), (3, "X"), (8, "Z")), 1.2)
    operator += QubitOperator(((1, "Z"), (3, "Y"), (9, "Z")), 1.4j)
    res = operator * operator
    correct = QubitOperator((), 0.5**2 + 1.2**2 + 1.4j**2)
    correct += QubitOperator(((1, "Y"), (3, "Z")), 2j * 1j * 0.5 * 1.2)
    assert res == correct


def test_renormalize_error():
    operator = QubitOperator()
    with pytest.raises(ZeroDivisionError):
        operator.renormalize()


def test_renormalize():
    operator = QubitOperator(((1, "X"), (3, "Y"), (8, "Z")), 1)
    operator += QubitOperator(((2, "Z"), (3, "Y")), 1)
    operator.renormalize()
    for term in operator.terms:
        assert operator.terms[term] == pytest.approx(1 / numpy.sqrt(2.0))
    assert operator.induced_norm(2) == pytest.approx(1.0)


def test_get_operators_empty():
    """Tests get_operators() with zero operator."""
    operator = QubitOperator.zero()

    operators = list(operator.get_operators())
    assert operators == []


def test_get_operators_one():
    """Tests get_operators() with an operator with a single term."""
    operator = QubitOperator(((1, "X"), (3, "Y"), (8, "Z")), 1)

    operators = list(operator.get_operators())
    assert operators == [operator]


def test_get_operators():
    """Tests get_operators() with an operator with two terms."""
    operator_00 = QubitOperator(((1, "X"), (3, "Y"), (8, "Z")), 1)
    operator_01 = QubitOperator(((2, "Z"), (3, "Y")), 1)
    sum_operator = operator_00 + operator_01

    operators = list(sum_operator.get_operators())
    assert operators in [[operator_00, operator_01], [operator_01, operator_00]]


def test_get_operator_groups_empty():
    """Tests get_operator_groups() with zero operator."""
    operator = QubitOperator.zero()

    operators = list(operator.get_operator_groups(1))
    assert operators == []


# Utility functions.
def generate_operator(begin, end):
    """Returns a sum of Z operators at qubit [begin, end)."""
    operator = QubitOperator.zero()
    for i in range(begin, end):
        operator += QubitOperator(((i, "Z"),), 1)
    return operator


def check_length(operators, lens):
    """Checks length operator is the same to lens."""
    if len(operators) != len(lens):
        return False

    for operator, length in zip(operators, lens):
        if len(operator.terms) != length:
            return False
    return True


def check_sum(operators, operator):
    """Checks sum of operators matches to operator."""
    return QubitOperator.accumulate(operators) == operator


# Tests for get_operator_groups().
def test_get_operator_groups_zero():
    """Tests get_operator_groups() with one group."""
    operator = generate_operator(0, 20)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        operator_groups = list(operator.get_operator_groups(0))

    # Using 1 group instead.
    assert check_length(operator_groups, [20])
    assert check_sum(operator_groups, operator)

    # Not using 0 groups.
    assert not check_length(operator_groups, [])
    assert not check_length(operator_groups, [0])
    assert not check_sum(operator_groups, operator * 2)


def test_get_operator_groups_one():
    """Tests get_operator_groups() with one group."""
    operator = generate_operator(0, 20)
    operator_groups = list(operator.get_operator_groups(1))

    assert check_length(operator_groups, [20])
    assert check_sum(operator_groups, operator)


def test_get_operator_groups_two():
    """Tests get_operator_groups() with two groups."""
    operator = generate_operator(0, 20)
    operator_groups = list(operator.get_operator_groups(2))

    assert check_length(operator_groups, [10, 10])
    assert check_sum(operator_groups, operator)


def test_get_operator_groups_three():
    """Tests get_operator_groups() with two groups."""
    operator = generate_operator(0, 20)
    operator_groups = list(operator.get_operator_groups(3))

    assert check_length(operator_groups, [7, 7, 6])
    assert check_sum(operator_groups, operator)


def test_get_operator_groups_six():
    """Tests get_operator_groups() with two groups."""
    operator = generate_operator(0, 20)
    operator_groups = list(operator.get_operator_groups(6))

    assert check_length(operator_groups, [4, 4, 3, 3, 3, 3])
    assert check_sum(operator_groups, operator)
